#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import json
import textwrap
from collections.abc import Iterator
from contextlib import contextmanager
from dataclasses import dataclass
from uuid import UUID

from evo.common import RequestMethod
from evo.common.exceptions import UnknownResponseError
from evo.common.test_tools import ORG as TEST_ORG
from evo.common.test_tools import MockResponse, TestWithConnector
from parameterized import parameterized
from pydantic import BaseModel

from data import load_test_data
from evo.compute import JobClient, JobProgress, JobStatusEnum
from evo.compute.exceptions import JobError, JobPendingError

TEST_TOPIC = "test"
TEST_TASK = "job-client"
TEST_JOB_ID = UUID(int=1234)


class PydanticResult(BaseModel):
    foo: str
    baz: int


@dataclass(frozen=True, kw_only=True)
class DataclassResult:
    foo: str
    baz: int


class TestJobClient(TestWithConnector):
    def setUp(self) -> None:
        super().setUp()
        self.job = JobClient(
            connector=self.connector,
            org_id=TEST_ORG.id,
            topic=TEST_TOPIC,
            task=TEST_TASK,
            job_id=TEST_JOB_ID,
        )

    @property
    def task_path(self) -> str:
        return f"/compute/orgs/{TEST_ORG.id}/{self.job.topic}/{self.job.task}"

    @property
    def job_url(self) -> str:
        return self.connector.base_url.rstrip("/") + self.task_path + f"/{self.job.id}/status"

    def test_id(self) -> None:
        """Test that the job ID is exposed."""
        self.assertEqual(TEST_JOB_ID, self.job.id)

    def test_topic(self) -> None:
        """Test that the topic is exposed."""
        self.assertEqual(TEST_TOPIC, self.job.topic)

    def test_task(self) -> None:
        """Test that the task is exposed."""
        self.assertEqual(TEST_TASK, self.job.task)

    def test_url(self) -> None:
        """Test that the job URL is correctly constructed."""
        self.assertEqual(self.job_url, self.job.url)

    def test_repr(self) -> None:
        """Test that the job is represented as expected."""
        self.assertEqual(self.job_url, repr(self.job))

    def assert_jobs_equal(self, job1: JobClient, job2: JobClient) -> None:
        """Assert that two jobs are equal, but not the same object."""
        self.assertIsNot(job1, job2)
        self.assertEqual(job1.id, job2.id)
        self.assertEqual(job1.topic, job2.topic)
        self.assertEqual(job1.task, job2.task)
        self.assertEqual(job1.url, job2.url)

    def test_from_url(self) -> None:
        """Test that a job can be constructed from a URL."""
        job = JobClient.from_url(self.connector, self.job_url)
        self.assert_jobs_equal(self.job, job)

    async def test_submit(self) -> None:
        """Test that a job can be submitted."""
        with self.transport.set_http_response(status_code=303, headers={"Location": self.job_url}):
            job = await JobClient.submit(
                connector=self.connector,
                org_id=TEST_ORG.id,
                topic=TEST_TOPIC,
                task=TEST_TASK,
                parameters={"foo": "bar"},
            )
        self.assert_request_made(
            RequestMethod.POST,
            self.task_path,
            headers={"Content-Type": "application/json"},
            body={"parameters": {"foo": "bar"}},
        )
        self.assert_jobs_equal(self.job, job)

    @contextmanager
    def set_status_response(self, http_status: int, data_file: str) -> Iterator[JobProgress]:
        """Set the response for a status request and yield the expected status."""
        response_data = load_test_data(data_file)
        response_data.pop("results", None)

        if response_data.get("error") is not None:
            error = JobError(
                status=response_data["error"].get("status"),
                reason=None,
                content=response_data["error"],
                headers=None,
            )
        else:
            error = None
        expected_status = JobProgress(
            status=JobStatusEnum(response_data["status"]),
            progress=response_data.get("progress"),
            message=response_data.get("message"),
            error=error,
        )

        response_json = json.dumps(response_data)
        with self.transport.set_http_response(
            status_code=http_status, content=response_json, headers={"Content-Type": "application/json"}
        ):
            yield expected_status

            self.assert_request_made(
                RequestMethod.GET,
                self.task_path + f"/{self.job.id}/status",
                headers={"Accept": "application/json"},
            )

    @parameterized.expand(
        [
            ("requested", 202, "job-response-requested.json"),
            ("in_progress", 202, "job-response-in-progress.json"),
            ("cancelling", 202, "job-response-cancelling.json"),
            ("cancelled", 200, "job-response-cancelled.json"),
            ("failed", 200, "job-response-failed.json"),
            ("succeeded", 200, "job-response-succeeded.json"),
        ]
    )
    async def test_get_status(self, _label: str, http_status: int, data_file: str) -> None:
        """Test retrieving the status of a job."""
        with self.set_status_response(http_status, data_file) as expected_status:
            reported_status = await self.job.get_status()
        self.assertEqual(expected_status.status, reported_status.status)
        self.assertEqual(expected_status.progress, reported_status.progress)
        self.assertEqual(expected_status.message, reported_status.message)
        # JobError does not implement __eq__, so compare their string representations.
        self.assertEqual(str(expected_status.error), str(reported_status.error))

    def test_job_progress_failed_str(self) -> None:
        """Test the string representation of a failed JobProgress."""
        progress = JobProgress(
            status=JobStatusEnum.failed,
            progress=None,
            message="Job failed due to error",
            error=JobError(
                status=422,
                reason=None,
                content={
                    "title": "Unprocessable Entity",
                    "type": "https://example.com/errors/422",
                    "detail": "Invalid parameters",
                },
                headers=None,
            ),
        )
        expected_str = """\
        [failed] > Job failed due to error
        Error: (422)
        Type: https://example.com/errors/422
        Title: Unprocessable Entity
        Detail: Invalid parameters"""
        self.assertEqual(textwrap.dedent(expected_str), str(progress))

    @contextmanager
    def set_result_response(self, http_status: int, data_file: str) -> Iterator[dict | None]:
        """Set the response for a result request and yield the expected results."""
        response_data = load_test_data(data_file)
        response_json = json.dumps(response_data)
        with self.transport.set_http_response(
            status_code=http_status, content=response_json, headers={"Content-Type": "application/json"}
        ):
            yield response_data.get("results")
            self.assert_request_made(
                RequestMethod.GET,
                self.task_path + f"/{self.job.id}",
                headers={"Accept": "application/json"},
            )

    @parameterized.expand(
        [
            ("job requested", 202, "job-response-requested.json", JobPendingError),
            ("job in progress", 202, "job-response-in-progress.json", JobPendingError),
            ("job cancelling", 202, "job-response-cancelling.json", JobPendingError),
            ("job failed", 200, "job-response-failed.json", JobError),
            # It is unclear what the actual expected response is for a cancelled job.
            # For now we assume there are no results and no error.
            ("job cancelled", 200, "job-response-cancelled.json", UnknownResponseError),
        ]
    )
    async def test_get_results_errors(
        self, _label: str, http_status: int, data_file: str, exc_type: type[Exception]
    ) -> None:
        """Test that errors are raised when requesting results for a job in an invalid state."""
        with self.set_result_response(http_status, data_file):
            with self.assertRaises(exc_type):
                await self.job.get_results()

    async def test_get_results(self) -> None:
        """Test that results can be retrieved for a job."""
        with self.set_result_response(200, "job-response-succeeded.json") as expected_results:
            results = await self.job.get_results()
        self.assertEqual(expected_results, results)

    async def test_get_results_twice(self) -> None:
        with self.set_result_response(200, "job-response-succeeded.json") as expected_results:
            first_results = await self.job.get_results()
        self.assertEqual(expected_results, first_results)

        self.transport.reset_mock()
        with self.transport.set_http_response(status_code=500):  # Should not be called.
            second_results = await self.job.get_results()
        self.transport.assert_no_requests()
        self.assertEqual(first_results, second_results)
        self.assertIsNot(first_results, second_results)  # Ensure a new object is returned.

    async def test_get_results_as_pydantic(self) -> None:
        job = JobClient.from_url(self.connector, self.job_url, result_type=PydanticResult)

        with self.set_result_response(200, "job-response-succeeded.json") as expected_results:
            results = await job.get_results()

        expected_results = PydanticResult.model_validate(expected_results)
        self.assertIsInstance(results, PydanticResult)
        self.assertEqual(expected_results, results)

    async def test_get_results_as_dataclass(self) -> None:
        job = JobClient.from_url(self.connector, self.job_url, result_type=DataclassResult)

        with self.set_result_response(200, "job-response-succeeded.json") as expected_results:
            results = await job.get_results()

        expected_results = DataclassResult(**expected_results)
        self.assertIsInstance(results, DataclassResult)
        self.assertEqual(expected_results, results)

    async def test_cancel(self) -> None:
        """Test that a job can be cancelled."""
        with self.transport.set_http_response(status_code=204):
            await self.job.cancel()
        self.assert_request_made(
            RequestMethod.DELETE,
            self.task_path + f"/{self.job.id}",
        )

    @contextmanager
    def set_job_states(self, *data_files: str) -> Iterator[dict | None]:
        """Set the responses for a sequence of job status requests and yield the final results."""
        all_responses = [
            MockResponse(
                status_code=202,
                headers={"Content-Type": "application/json"},
                content=json.dumps(load_test_data(data_file)),
            )
            for data_file in data_files
        ]
        final_response = load_test_data(data_files[-1])
        all_responses.append(
            MockResponse(
                status_code=200,
                headers={"Content-Type": "application/json"},
                content=json.dumps(final_response),
            )
        )
        old_side_effect = self.transport.request.side_effect
        self.transport.request.side_effect = all_responses
        try:
            yield final_response.get("results")
        finally:
            self.transport.request.side_effect = old_side_effect

    async def test_wait_for_result(self) -> None:
        """Test that a job can be waited on for results."""
        with self.set_job_states(
            "job-response-requested.json",
            "job-response-in-progress.json",
            "job-response-succeeded.json",
        ) as expected_results:
            results = await self.job.wait_for_results(polling_interval_seconds=0.0)

        self.assert_any_request_made(
            RequestMethod.GET,
            self.task_path + f"/{self.job.id}/status",
            headers={"Accept": "application/json"},
        )
        self.assert_any_request_made(
            RequestMethod.GET,
            self.task_path + f"/{self.job.id}",
            headers={"Accept": "application/json"},
        )

        self.assertEqual(expected_results, results)

    async def test_wait_for_result_failed(self) -> None:
        """Test waiting for a job that fails."""
        with self.set_job_states(
            "job-response-requested.json",
            "job-response-in-progress.json",
            "job-response-failed.json",
        ):
            with self.assertRaises(JobError):
                await self.job.wait_for_results(polling_interval_seconds=0.0)

        self.assert_any_request_made(
            RequestMethod.GET,
            self.task_path + f"/{self.job.id}/status",
            headers={"Accept": "application/json"},
        )
        self.assert_any_request_made(
            RequestMethod.GET,
            self.task_path + f"/{self.job.id}",
            headers={"Accept": "application/json"},
        )

    async def test_wait_for_result_cancelled(self) -> None:
        """Test waiting for a job that gets cancelled."""
        with self.set_job_states(
            "job-response-requested.json",
            "job-response-in-progress.json",
            "job-response-cancelling.json",
            "job-response-cancelled.json",
        ):
            with self.assertRaises(UnknownResponseError):
                await self.job.wait_for_results(polling_interval_seconds=0.0)

        self.assert_any_request_made(
            RequestMethod.GET,
            self.task_path + f"/{self.job.id}/status",
            headers={"Accept": "application/json"},
        )
        self.assert_any_request_made(
            RequestMethod.GET,
            self.task_path + f"/{self.job.id}",
            headers={"Accept": "application/json"},
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import enum
import re
from dataclasses import dataclass
from datetime import datetime
from typing import Protocol
from urllib.parse import parse_qs, urlparse
from uuid import UUID

from evo.common import Environment, ResourceMetadata
from evo.workspaces import ServiceUser

from .exceptions import SchemaIDFormatError

__all__ = [
    "ObjectMetadata",
    "ObjectOrderByEnum",
    "ObjectReference",
    "ObjectSchema",
    "ObjectVersion",
    "SchemaVersion",
    "Stage",
]


class ObjectOrderByEnum(str, enum.Enum):
    author = "author"
    created_at = "created_at"
    created_by = "created_by"
    deleted_at = "deleted_at"
    modified_at = "modified_at"
    modified_by = "modified_by"
    object_name = "object_name"


class ObjectReference(str):
    """A structured URL reference to a geoscience object, optionally including a version ID.

    Geoscience Object URL references are the fully qualified HTTPS URLs used to access objects in the
    Geoscience Object API. The URL may follow the path or UUID format, and may optionally include a version ID.

    In most cases, UUID-based references are preferred, as they are immutable and unambiguous. However, path-based references
    can be useful in scenarios where the object ID is not known, such as when creating new objects or when working with
    objects in a more human-readable way.
    """

    _RE_PATH = re.compile(
        r"""
        ^/geoscience-object
        /orgs/(?P<org_id>[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})
        /workspaces/(?P<workspace_id>[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})
        /objects
        (?:
            /path/(?P<object_path>[^?]+) | /(?P<object_id>[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12})
        )$
        """,
        re.IGNORECASE | re.VERBOSE,
    )

    hub_url: str
    """The base URL of the Evo Hub."""

    org_id: UUID
    """The ID of the Evo Organization the object belongs to."""

    workspace_id: UUID
    """The ID of the Evo Workspace the object belongs to."""

    object_id: UUID | None
    """The UUID of the object, if specified in the URL."""

    object_path: str | None
    """The path of the object, if specified in the URL."""

    version_id: str | None
    """The version ID of the object, if specified in the URL."""

    def __new__(cls, value: str) -> ObjectReference:
        inst = str.__new__(cls, value)

        parsed = urlparse(value)
        if parsed.scheme != "https":
            raise ValueError("Reference must be a valid HTTPS URL")

        inst.hub_url = f"{parsed.scheme}://{parsed.netloc}/"

        if match := cls._RE_PATH.fullmatch(parsed.path):
            inst.org_id = UUID(match.group("org_id"))
            inst.workspace_id = UUID(match.group("workspace_id"))

            if match.group("object_id"):
                inst.object_id = UUID(match.group("object_id"))
                inst.object_path = None
            else:
                inst.object_id = None
                inst.object_path = match.group("object_path").lstrip("/")
        else:
            raise ValueError("Reference path is not valid")

        query_params = parse_qs(parsed.query)
        inst.version_id = query_params.get("version", [None])[0]
        return inst

    @property
    def environment(self) -> Environment:
        return Environment(hub_url=self.hub_url, org_id=self.org_id, workspace_id=self.workspace_id)

    @staticmethod
    def new(
        environment: Environment,
        object_id: UUID | None = None,
        object_path: str | None = None,
        version_id: str | None = None,
    ) -> ObjectReference:
        """Create a new ObjectReference from its components.

        Either object_id or object_path must be provided, but not both.

        :param environment: The Evo environment the object belongs to.
        :param object_id: The UUID of the object, if known.
        :param object_path: The path of the object, if known.
        :param version_id: The version ID of the object, if known.

        :returns: A new ObjectReference instance.

        :raises ValueError: If neither or both of object_id and object_path are provided.
        """
        if object_id is None and object_path is None:
            raise ValueError("Either object_id or object_path must be provided")
        if object_id is not None and object_path is not None:
            raise ValueError("Only one of object_id or object_path can be provided")

        if object_id is not None:
            path = (
                f"geoscience-object/orgs/{environment.org_id}/workspaces/{environment.workspace_id}/objects/{object_id}"
            )
        else:
            path = f"geoscience-object/orgs/{environment.org_id}/workspaces/{environment.workspace_id}/objects/path/{object_path.lstrip('/')}"

        if version_id is not None:
            path += f"?version={version_id}"

        return ObjectReference(f"{environment.hub_url.rstrip('/')}/{path}")


@dataclass(frozen=True, kw_only=True)
class ObjectMetadata(ResourceMetadata):
    """Metadata about a geoscience object."""

    parent: str
    """The parent path of the object."""

    schema_id: ObjectSchema
    """The geoscience object schema."""

    version_id: str
    """An arbitrary identifier for the object version."""

    modified_at: datetime
    """The date and time when the object was last modified."""

    modified_by: ServiceUser | None
    """The user who last modified the object."""

    stage: Stage | None
    """The stage of the object, if available."""

    @property
    def path(self) -> str:
        """The full path of the object, formed by joining the parent and name, separated by a slash ('/')."""
        return f"{self.parent}/{self.name}"

    @property
    def url(self) -> ObjectReference:
        """The url of the object."""
        return ObjectReference.new(
            environment=self.environment,
            object_id=self.id,
            version_id=self.version_id,
        )


@dataclass(frozen=True, kw_only=True)
class OrgObjectMetadata(ResourceMetadata):
    """Metadata about a geoscience object in an organization listing."""

    workspace_id: UUID
    """The ID of the workspace that contains the object."""

    workspace_name: str | None
    """The name of the workspace that contains the object."""

    schema_id: ObjectSchema
    """The geoscience object schema."""

    modified_at: datetime
    """The date and time when the object was last modified."""

    modified_by: ServiceUser | None
    """The user who last modified the object."""

    stage: Stage | None
    """The stage of the object, if available."""

    @property
    def url(self) -> str:
        """The url of the object."""
        return ObjectReference.new(
            environment=self.environment,
            object_id=self.id,
        )


@dataclass(frozen=True, kw_only=True)
class ObjectVersion:
    """Represents a version of an object."""

    version_id: str
    """Used by the service to identify a unique resource version."""

    created_at: datetime
    """A UTC timestamp representing when the version was uploaded to the service."""

    created_by: ServiceUser | None
    """The user that uploaded the version."""

    stage: Stage | None
    """The stage associated with this version, if applicable."""


class _StageModel(Protocol):
    stage_id: UUID
    name: str


@dataclass(frozen=True, kw_only=True)
class Stage:
    """Metadata about a stage"""

    id: UUID
    """The stage UUID."""

    name: str
    """The stage name."""

    @classmethod
    def from_model(cls, model: _StageModel) -> Stage:
        """Create a new instance from an instance of a generated model."""
        return cls(id=model.stage_id, name=model.name)


@dataclass(frozen=True, order=True)
class SchemaVersion:
    major: int
    minor: int
    patch: int

    @classmethod
    def from_str(cls, string: str) -> SchemaVersion:
        return cls(*[int(i) for i in string.split(".")])

    def __str__(self) -> str:
        return f"{self.major}.{self.minor}.{self.patch}"


@dataclass(frozen=True)
class ObjectSchema:
    root_classification: str
    sub_classification: str
    version: SchemaVersion

    @property
    def classification(self) -> str:
        return f"{self.root_classification}/{self.sub_classification}"

    @classmethod
    def from_id(cls, schema_id: str) -> ObjectSchema:
        schema_components = re.match(
            r"/(?P<root>[-\w]+)/(?P<sub>[-\w]+)/(?P<version>\d+\.\d+\.\d+)/(?P=sub)\.schema\.json",
            schema_id,
        )
        if schema_components is None:
            raise SchemaIDFormatError(f"Could not parse schema id: '{schema_id}'")

        return cls(
            root_classification=schema_components.group("root"),
            sub_classification=schema_components.group("sub"),
            version=SchemaVersion.from_str(schema_components.group("version")),
        )

    def __str__(self) -> str:
        return f"/{self.classification}/{self.version}/{self.sub_classification}.schema.json"



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

try:
    import pyarrow  # noqa: F401
except ImportError:
    raise ImportError("The 'pyarrow' package is required to use ParquetLoader") from None

from .loader import ParquetDownloader, ParquetLoader
from .types import ArrayTableInfo, AttributeInfo, CategoryInfo, LookupTableInfo, TableInfo

__all__ = [
    "ArrayTableInfo",
    "AttributeInfo",
    "CategoryInfo",
    "LookupTableInfo",
    "ParquetDownloader",
    "ParquetLoader",
    "TableInfo",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Compute Task API
=============

Normally this code would be generated from an OpenAPI specification. However, the API specification for
the Compute Task API is new and not yet finalized. Therefore, this code is written against the intended API,
with some customisation as needed to make it work.

The client implementation should still abstract these details away from the user, so that the user can interact
with the API in a more user-friendly way, and so that this implementation can be changed without affecting the user.

This code is based on the OpenAPI specification for Compute Task API.
API version: 0.0.1
"""

# Import endpoint apis.
from .api import JobApi, TaskApi

__all__ = [
    "JobApi",
    "TaskApi",
]



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from __future__ import annotations

import asyncio
import uuid
from uuid import UUID

from evo import logging
from evo.common import APIConnector, BaseAPIClient, Environment, HealthCheckType, ICache, ServiceHealth
from evo.common.data import ServiceUser
from evo.common.utils import get_service_health

from ._types import Table
from ._utils import convert_dtype, extract_payload
from .data import BaseGridDefinition, BlockModel, RegularGridDefinition, Version
from .endpoints import models
from .endpoints.api import ColumnOperationsApi, JobsApi, OperationsApi, VersionsApi
from .endpoints.models import (
    AnyUrl,
    BBox,
    BBoxXYZ,
    BlockSize,
    ColumnHeaderType,
    CreateData,
    GeometryColumns,
    JobErrorPayload,
    JobResponse,
    JobStatus,
    Location,
    OutputOptionsParquet,
    QueryCriteria,
    QueryDownload,
    Rotation,
    RotationAxis,
    Size3D,
    SizeOptionsRegular,
)
from .exceptions import CacheNotConfiguredException, JobFailedException, MissingColumnInTable
from .io import BlockModelDownload, BlockModelUpload, get_cache_location_for_upload

logger = logging.getLogger("blockmodel.client")

__all__ = [
    "BlockModelAPIClient",
]


def _job_id_from_url(job_url: AnyUrl) -> UUID:
    job_id = job_url.path.split("/")[-1]
    return UUID(job_id)


def _version_from_model(version: models.Version) -> Version:
    return Version(
        bm_uuid=version.bm_uuid,
        version_id=version.version_id,
        version_uuid=version.version_uuid,
        created_at=version.created_at,
        created_by=ServiceUser.from_model(version.created_by),
        comment=version.comment,
        bbox=version.bbox,
        base_version_id=version.base_version_id,
        parent_version_id=version.parent_version_id,
        columns=version.mapping.columns,
        geoscience_version_id=version.geoscience_version_id,
    )


_GEOMETRY_COLUMNS = {"i", "j", "k", "x", "y", "z"}


class BlockModelAPIClient(BaseAPIClient):
    def __init__(self, environment: Environment, connector: APIConnector, cache: ICache | None = None) -> None:
        """
        Constructor for the Block Model Service client.

        Some methods need a cache to store temporary files. If you want to use these methods, you must provide a cache.

        :param environment: The environment object.
        :param connector: The connector object.
        :param cache: The cache to use for storing temporary files.
        """
        super().__init__(environment, connector)
        self._versions_api = VersionsApi(connector)
        self._jobs_api = JobsApi(connector)
        self._operations_api = OperationsApi(connector)
        self._column_operations_api = ColumnOperationsApi(connector)
        self._cache = cache

    async def get_service_health(self, check_type: HealthCheckType = HealthCheckType.FULL) -> ServiceHealth:
        """Get the health of the service.

        :param check_type: The type of health check to perform.

        :return: A ServiceHealth object.

        :raises EvoAPIException: If the API returns an unexpected status code.
        :raises ClientValueError: If the response is not a valid service health check response.
        """
        return await get_service_health(self._connector, "blockmodel", check_type=check_type)

    def _bm_from_model(self, model: models.BlockModel | models.BlockModelAndJobURL) -> BlockModel:
        match model.size_options:
            case SizeOptionsRegular(n_blocks=n_blocks, block_size=block_size):
                grid_definition = RegularGridDefinition(
                    model_origin=[model.model_origin.x, model.model_origin.y, model.model_origin.z],
                    rotations=[(rotation.axis, rotation.angle) for rotation in model.block_rotation],
                    n_blocks=[n_blocks.nx, n_blocks.ny, n_blocks.nz],
                    block_size=[block_size.x, block_size.y, block_size.z],
                )
            case _:
                raise NotImplementedError("Only regular models are supported at the moment")

        return BlockModel(
            environment=self._environment,
            id=model.bm_uuid,
            name=model.name,
            created_at=model.created_at,
            created_by=ServiceUser.from_model(model.created_by),
            description=model.description,
            grid_definition=grid_definition,
            coordinate_reference_system=model.coordinate_reference_system,
            size_unit_id=model.size_unit_id,
            bbox=model.bbox,
            last_updated_at=model.last_updated_at,
            last_updated_by=ServiceUser.from_model(model.last_updated_by),
            geoscience_object_id=model.geoscience_object_id,
        )

    async def _poll_job_url(self, bm_id: UUID, job_id: UUID) -> JobResponse:
        while True:
            response = await self._jobs_api.get_job_status(
                job_id=str(job_id),
                workspace_id=str(self._environment.workspace_id),
                org_id=str(self._environment.org_id),
                bm_id=str(bm_id),
            )
            if response.job_status == JobStatus.COMPLETE:
                return response
            elif response.job_status == JobStatus.FAILED:
                payload = extract_payload(job_id, response, JobErrorPayload)
                raise JobFailedException(job_id, payload)
            await asyncio.sleep(1)

    async def _create_block_model(
        self,
        name: str,
        grid_definition: BaseGridDefinition,
        description: str | None = None,
        object_path: str | None = None,
        coordinate_reference_system: str | None = None,
        size_unit_id: str | None = None,
    ):
        match grid_definition:
            case RegularGridDefinition(n_blocks=n_blocks, block_size=block_size):
                size_option = SizeOptionsRegular(
                    model_type="regular",
                    n_blocks=Size3D(nx=n_blocks[0], ny=n_blocks[1], nz=n_blocks[2]),
                    block_size=BlockSize(x=block_size[0], y=block_size[1], z=block_size[2]),
                )
            case _:
                raise NotImplementedError("Only regular models are supported at the moment")
        create_result = await self._operations_api.create_block_model(
            workspace_id=str(self._environment.workspace_id),
            org_id=str(self._environment.org_id),
            create_data=CreateData(
                name=name,
                description=description,
                object_path=object_path,
                coordinate_reference_system=coordinate_reference_system,
                size_unit_id=size_unit_id,
                model_origin=Location(
                    x=grid_definition.model_origin[0],
                    y=grid_definition.model_origin[1],
                    z=grid_definition.model_origin[2],
                ),
                block_rotation=[
                    Rotation(axis=RotationAxis(axis), angle=angle) for axis, angle in grid_definition.rotations
                ],
                size_options=size_option,
            ),
        )
        job_id = _job_id_from_url(create_result.job_url)
        job_status = await self._poll_job_url(create_result.bm_uuid, job_id)
        version = extract_payload(job_id, job_status, models.Version)
        return create_result, _version_from_model(version)

    async def _upload_data(self, bm_id: uuid.UUID, job_id: uuid.UUID, upload_url: str, data: Table) -> models.Version:
        """Upload data to a block model service, marks the upload as complete and waits for the job to complete."""
        # Write the data to a temporary file
        import pyarrow.parquet

        cache_location = get_cache_location_for_upload(self._cache, self._environment, job_id)
        pyarrow.parquet.write_table(data, cache_location)

        # Upload the data
        upload = BlockModelUpload(self._connector, self._environment, bm_id, job_id, upload_url)
        await upload.upload_from_path(cache_location, self._connector.transport)

        # Notify the service that the upload is complete
        await self._column_operations_api.notify_upload_complete(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            bm_id=str(bm_id),
            job_id=str(job_id),
        )

        # Poll the job URL until it is complete
        job_status = await self._poll_job_url(bm_id, job_id)
        return extract_payload(job_id, job_status, models.Version)

    async def create_block_model(
        self,
        name: str,
        grid_definition: BaseGridDefinition,
        description: str | None = None,
        object_path: str | None = None,
        coordinate_reference_system: str | None = None,
        size_unit_id: str | None = None,
        initial_data: Table | None = None,
        units: dict[str, str] | None = None,
    ) -> tuple[BlockModel, Version]:
        r"""Create a block model.

        Optionally, takes initial data to populate the block model with. Units for the columns within the initial data
        can be provided in the `units` dictionary. This requires the `pyarrow` package to be installed, and the 'cache'
        parameter to be set in the constructor.

        If `initial_data` is provided, this method will wait for the initial data to be successfully uploaded and
        processed before returning. This then returns both the block model and the version created from the initial data
        update.

        Otherwise, if `initial_data` is not provided, this waits for the block model creation job to complete before
        returning. This then returns both the block model and the initial block model version.

        :param name: Name of the block model. This may not contain `/` nor `\`.
        :param grid_definition: Definition of the block model grid.
        :param description: Description of the block model.
        :param object_path: Path of the folder in Geoscience Object Service to create the reference object in.
        :param coordinate_reference_system: Coordinate reference system used in the block model.
        :param size_unit_id: Unit ID denoting the length unit used for the block model's blocks.
        :param initial_data: The initial data to populate the block model with.
        :param units: A dictionary mapping column names within `initial_data` to units.
        :return: A tuple containing the created block model and the version of the block model.
        """
        if units is not None and initial_data is None:
            raise ValueError("units can only be provided if initial_data is provided")
        if initial_data is not None and self._cache is None:
            raise CacheNotConfiguredException(
                "Cache must be configured to use this method. Please set the 'cache' parameter in the constructor."
            )
        create_result, version = await self._create_block_model(
            name, grid_definition, description, object_path, coordinate_reference_system, size_unit_id
        )

        if initial_data is not None:
            version = await self.add_new_columns(create_result.bm_uuid, initial_data, units)
        return self._bm_from_model(create_result), version

    async def add_new_columns(
        self,
        bm_id: UUID,
        data: Table,
        units: dict[str, str] | None = None,
    ):
        """Add new columns to an existing block model.

        Units for the columns can be provided in the `units` dictionary.

        This method requires the `pyarrow` package to be installed, and the 'cache' parameter to be set in the constructor.

        :param bm_id: The ID of the block model to add columns to.
        :param data: The data containing the new columns to add.
        :param units: A dictionary mapping column names within `data` to units.
        :raises CacheNotConfiguredException: If the cache is not configured.
        :return: The new version of the block model with the added columns.
        """
        if self._cache is None:
            raise CacheNotConfiguredException(
                "Cache must be configured to use this method. Please set the 'cache' parameter in the constructor."
            )

        schema = data.schema
        if units is None:
            units = {}
        columns = models.UpdateColumnsLiteInput(
            new=[
                models.ColumnLite(title=name, data_type=convert_dtype(data_type), unit_id=units.get(name))
                for name, data_type in zip(schema.names, schema.types)
                if name not in _GEOMETRY_COLUMNS
            ],
            update=[],
            delete=[],
            rename=[],
        )
        update_response = await self._column_operations_api.update_block_model_from_latest_version(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            bm_id=str(bm_id),
            update_data_lite_input=models.UpdateDataLiteInput(
                columns=columns,
                update_type=models.UpdateType.replace,
            ),
        )
        version = await self._upload_data(bm_id, update_response.job_uuid, str(update_response.upload_url), data)
        return _version_from_model(version)

    async def update_block_model_columns(
        self,
        bm_id: UUID,
        data: Table,
        new_columns: list[str],
        update_columns: set[str] | None = None,
        delete_columns: set[str] | None = None,
        units: dict[str, str] | None = None,
    ) -> Version:
        """Add, update, or delete block model columns.

        Units for the columns can be provided in the `units` dictionary.

        This method requires the `pyarrow` package to be installed, and the 'cache' parameter to be set in the constructor.

        :param bm_id: The ID of the block model to add columns to.
        :param data: The data containing the new columns to add.
        :param new_columns: A list of new column names to add to the block model.
        :param update_columns: A set of column names to update in the block model.
        :param delete_columns: A set of column names to delete from the block model.
        :param units: A dictionary mapping column names within `data` to units.
        :raises CacheNotConfiguredException: If the cache is not configured.
        :return: The new version of the block model with the added columns.
        """
        if self._cache is None:
            raise CacheNotConfiguredException(
                "Cache must be configured to use this method. Please set the 'cache' parameter in the constructor."
            )

        schema = data.schema
        if units is None:
            units = {}
        data_type_map = {name: data_type for name, data_type in zip(schema.names, schema.types)}

        if update_columns is None:
            update_columns = set()

        if delete_columns is None:
            delete_columns = set()

        # Check for any new or updated columns that are not in the data
        missing = (set(new_columns) | update_columns) - data_type_map.keys()
        if missing:
            raise MissingColumnInTable(f"Columns {missing} are not present in the provided table.")

        columns = models.UpdateColumnsLiteInput(
            new=[
                models.ColumnLite(
                    title=new_column, data_type=convert_dtype(data_type_map[new_column]), unit_id=units.get(new_column)
                )
                for new_column in new_columns
            ],
            update=list(update_columns),
            delete=list(delete_columns),
            rename=[],
        )
        update_response = await self._column_operations_api.update_block_model_from_latest_version(
            org_id=str(self._environment.org_id),
            workspace_id=str(self._environment.workspace_id),
            bm_id=str(bm_id),
            update_data_lite_input=models.UpdateDataLiteInput(
                columns=columns,
                update_type=models.UpdateType.replace,
            ),
        )
        version = await self._upload_data(bm_id, update_response.job_uuid, str(update_response.upload_url), data)
        return _version_from_model(version)

    async def query_block_model_as_table(
        self,
        bm_id: UUID,
        columns: list[str | UUID],
        bbox: BBox | BBoxXYZ | None = None,
        version_uuid: UUID | None = None,
        geometry_columns: GeometryColumns = GeometryColumns.coordinates,
        column_headers: ColumnHeaderType = ColumnHeaderType.id,
        exclude_null_rows: bool = True,
    ) -> Table:
        """Query a block model and return the result as a PyArrow Table.

        This requires the `pyarrow` package to be installed, and the 'cache' parameter to be set in the constructor.

        :param bm_id: The ID of the block model to query.
        :param columns: The columns to query, can either be the title or the ID of the column.
        :param bbox: The bounding box to query, if None (the default) the entire block model is queried.
        :param version_uuid: The version UUID to query, if None (the default) the latest version is queried.
        :param geometry_columns: Whether rows in the returned table should include coordinates, or block indices of the
            block, that the row belongs to.
        :param column_headers: Whether the names of the columns in the returned column should be the title or the ID of
            the block model column.
        :param exclude_null_rows: Whether to exclude rows where all values are null within the queried columns.
        :return: The result as a PyArrow Table.
        :raises JobFailedException: If the job failed.
        """
        import pyarrow.parquet

        if self._cache is None:
            raise CacheNotConfiguredException(
                "Cache must be configured to use this method. Please set the 'cache' parameter in the constructor."
            )

        query_result = await self._versions_api.query_block_model_latest_as_post(
            bm_id=str(bm_id),
            workspace_id=str(self._environment.workspace_id),
            org_id=str(self._environment.org_id),
            query_criteria=QueryCriteria(
                columns=[str(c) for c in columns],  # Convert any UUIDs to strings
                bbox=bbox,
                version_uuid=version_uuid,
                geometry_columns=geometry_columns,
                output_options=OutputOptionsParquet(
                    file_format="parquet",
                    column_headers=column_headers,
                    exclude_null_rows=exclude_null_rows,
                ),
            ),
        )

        # Poll the job URL until it is complete
        job_id = _job_id_from_url(query_result.job_url)
        job = await self._poll_job_url(bm_id, job_id)
        payload = extract_payload(job_id, job, QueryDownload)

        # Download the result to a temporary file
        download = BlockModelDownload(
            self._connector, self._environment, query_result, job_id, str(payload.download_url)
        )
        path = await download.download_to_cache(self._cache, self._connector.transport)

        # Read the PyArrow Table from the temporary file
        return pyarrow.parquet.read_table(path)



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest

from evo.files.io import _INVALID_NAMES, _make_name_safe


class TestMakeNameSafe(unittest.TestCase):
    def test_invalid_characters(self):
        self.assertEqual(_make_name_safe("test<name>"), "test%3Cname%3E")
        self.assertEqual(_make_name_safe("test:name"), "test%3Aname")
        self.assertEqual(_make_name_safe("test/name"), "test%2Fname")
        self.assertEqual(_make_name_safe("test\\name"), "test%5Cname")
        self.assertEqual(_make_name_safe("test|name"), "test%7Cname")
        self.assertEqual(_make_name_safe("test?name"), "test%3Fname")
        self.assertEqual(_make_name_safe("test*name"), "test%2Aname")

    def test_invalid_trailing_characters(self):
        self.assertEqual(_make_name_safe("test "), "test%20")
        self.assertEqual(_make_name_safe("test."), "test%2E")

    def test_reserved_names(self):
        for name in _INVALID_NAMES:
            self.assertEqual(_make_name_safe(name), name + "_")
            self.assertEqual(_make_name_safe(name + ".txt"), name + "_.txt")

    def test_valid_names(self):
        self.assertEqual(_make_name_safe("test"), "test")
        self.assertEqual(_make_name_safe("test123"), "test123")
        self.assertEqual(_make_name_safe("test_name"), "test_name")

    def test_leading_trailing_special_characters(self):
        self.assertEqual(_make_name_safe("<test>"), "%3Ctest%3E")

    def test_leading_trailing_reserved_names(self):
        self.assertEqual(_make_name_safe("CONtestCON"), "CONtestCON")

    def test_mixed_valid_invalid_characters(self):
        self.assertEqual(_make_name_safe("te<st>na:me"), "te%3Cst%3Ena%3Ame")

    def test_empty_string(self):
        self.assertEqual(_make_name_safe(""), "")

    def test_only_special_characters(self):
        self.assertEqual(_make_name_safe('<>:"/\\|?*'), "%3C%3E%3A%22%2F%5C%7C%3F%2A")



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
import datetime
import json
from unittest import mock
from urllib.parse import quote as quote
from uuid import UUID

from dateutil.parser import parse as dateutil_parse

from data import load_test_data
from evo.common import HealthCheckType, Page, RequestMethod, ServiceUser
from evo.common.data import OrderByOperatorEnum
from evo.common.io.exceptions import DataNotFoundError
from evo.common.test_tools import MockResponse, TestWithConnector, TestWithStorage
from evo.objects import (
    ObjectAPIClient,
    ObjectDataDownload,
    ObjectDataUpload,
    ObjectMetadata,
    ObjectSchema,
    ObjectVersion,
    SchemaVersion,
)
from evo.objects.data import ObjectOrderByEnum, OrgObjectMetadata, Stage
from evo.objects.exceptions import ObjectAlreadyExistsError, ObjectUUIDError
from evo.objects.utils import ObjectDataClient
from helpers import NoImport, UnloadModule

EMPTY_CONTENT = '{"objects": [], "links": {"next": null, "prev": null}}'
MOCK_VERSION_CONTENT = json.dumps(load_test_data("list_versions.json"))
_MAX_UPLOAD_URLS = 32


class TestObjectAPIClient(TestWithConnector, TestWithStorage):
    def setUp(self) -> None:
        TestWithConnector.setUp(self)
        TestWithStorage.setUp(self)
        self.object_client = ObjectAPIClient(connector=self.connector, environment=self.environment)

    @property
    def instance_base_path(self) -> str:
        return f"geoscience-object/orgs/{self.environment.org_id}"

    @property
    def base_path(self) -> str:
        return f"{self.instance_base_path}/workspaces/{self.environment.workspace_id}"

    async def test_check_service_health(self) -> None:
        """Test service health check implementation"""
        with mock.patch("evo.objects.client.api_client.get_service_health", spec_set=True) as mock_get_service_health:
            await self.object_client.get_service_health()
        mock_get_service_health.assert_called_once_with(
            self.connector, "geoscience-object", check_type=HealthCheckType.FULL
        )

    async def test_list_objects_default_args(self) -> None:
        with self.transport.set_http_response(200, EMPTY_CONTENT, headers={"Content-Type": "application/json"}):
            page = await self.object_client.list_objects()
        self.assertIsInstance(page, Page)
        self.assertEqual([], page.items())
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects?limit=5000&offset=0",
            headers={"Accept": "application/json"},
        )

    async def test_list_objects_all_args(self) -> None:
        with self.transport.set_http_response(200, EMPTY_CONTENT, headers={"Content-Type": "application/json"}):
            page = await self.object_client.list_objects(limit=20)
        self.assertIsInstance(page, Page)
        self.assertEqual([], page.items())
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects?limit=20&offset=0",
            headers={"Accept": "application/json"},
        )

    async def test_list_objects(self) -> None:
        content_0 = load_test_data("list_objects_0.json")
        content_1 = load_test_data("list_objects_1.json")
        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        page_one = await self.object_client.list_objects(
            limit=2, order_by={ObjectOrderByEnum.created_at: OrderByOperatorEnum.asc}, schema_id=["test"], deleted=False
        )
        expected_items_page_one = [
            ObjectMetadata(
                environment=self.environment,
                id=UUID("00000000-0000-0000-0000-000000000002"),
                name="m.json",
                created_at=dateutil_parse("2020-01-01 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                parent="/A",
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                version_id="2",
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000999")),
            ),
            ObjectMetadata(
                environment=self.environment,
                id=UUID("00000000-0000-0000-0000-000000000003"),
                name="n.json",
                created_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                modified_at=dateutil_parse("2020-01-03 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                parent="/A",
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                version_id="1",
                stage=None,
            ),
        ]
        self.assertIsInstance(page_one, Page)
        self.assertEqual(expected_items_page_one, page_one.items())
        self.assertEqual(0, page_one.offset)
        self.assertEqual(2, page_one.limit)
        self.assertFalse(page_one.is_last)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects?limit=2&offset=0&deleted=False&order_by=asc%3Acreated_at&schema_id=test",
            headers={"Accept": "application/json"},
        )
        self.transport.request.reset_mock()

        page_two = await self.object_client.list_objects(offset=page_one.next_offset, limit=page_one.limit)
        expected_items_page_two = [
            ObjectMetadata(
                environment=self.environment,
                id=UUID("00000000-0000-0000-0000-000000000002"),
                name="o.json",
                created_at=dateutil_parse("2020-01-01 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                parent="/B",
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                version_id="3",
                stage=None,
            ),
        ]
        self.assertIsInstance(page_two, Page)
        self.assertEqual(expected_items_page_two, page_two.items())
        self.assertEqual(2, page_two.offset)
        self.assertEqual(2, page_two.limit)
        self.assertTrue(page_two.is_last)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects?limit=2&offset=2",
            headers={"Accept": "application/json"},
        )

    async def test_list_objects_for_instance(self) -> None:
        content_0 = load_test_data("list_objects_for_instance_0.json")
        content_1 = load_test_data("list_objects_for_instance_1.json")
        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        page_one = await self.object_client.list_objects_for_instance(
            limit=2, order_by={ObjectOrderByEnum.created_at: OrderByOperatorEnum.asc}, schema_id=["test"], deleted=False
        )
        expected_items_page_one = [
            OrgObjectMetadata(
                environment=self.environment,
                workspace_id=UUID("00000000-0000-0000-0000-00000000162e"),
                workspace_name="Test Workspace",
                id=UUID("00000000-0000-0000-0000-000000000002"),
                name="m.json",
                created_at=dateutil_parse("2020-01-01 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000747")),
            ),
            OrgObjectMetadata(
                environment=self.environment,
                workspace_id=UUID("00000000-0000-0000-0000-00000000162e"),
                workspace_name="Test Workspace",
                id=UUID("00000000-0000-0000-0000-000000000003"),
                name="n.json",
                created_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                modified_at=dateutil_parse("2020-01-03 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                stage=None,
            ),
        ]
        self.assertIsInstance(page_one, Page)
        for item in page_one:
            self.assertEqual(item.environment.workspace_id, item.workspace_id, "workspace_id should match environment")
        self.assertEqual(expected_items_page_one, page_one.items())
        self.assertEqual(0, page_one.offset)
        self.assertEqual(2, page_one.limit)
        self.assertFalse(page_one.is_last)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.instance_base_path}/objects?offset=0&limit=2&deleted=False&permitted_workspaces_only=True&order_by=asc%3Acreated_at&schema_id=test",
            headers={"Accept": "application/json"},
        )
        self.transport.request.reset_mock()

        page_two = await self.object_client.list_objects_for_instance(offset=page_one.next_offset, limit=page_one.limit)
        expected_items_page_two = [
            OrgObjectMetadata(
                environment=dataclasses.replace(
                    self.environment, workspace_id=UUID("00000000-0000-0000-0000-0000000004d2")
                ),
                workspace_id=UUID("00000000-0000-0000-0000-0000000004d2"),
                workspace_name="Test Workspace 2",
                id=UUID("00000000-0000-0000-0000-000000000002"),
                name="o.json",
                created_at=dateutil_parse("2020-01-01 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                stage=None,
            ),
        ]
        self.assertIsInstance(page_two, Page)
        for item in page_two:
            self.assertEqual(item.environment.workspace_id, item.workspace_id, "workspace_id should match environment")
        self.assertEqual(expected_items_page_two, page_two.items())
        self.assertEqual(2, page_two.offset)
        self.assertEqual(2, page_two.limit)
        self.assertTrue(page_two.is_last)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.instance_base_path}/objects?offset=2&limit=2&permitted_workspaces_only=True",
            headers={"Accept": "application/json"},
        )

    async def test_list_all_objects(self) -> None:
        content_0 = load_test_data("list_objects_0.json")
        content_1 = load_test_data("list_objects_1.json")
        responses = [
            MockResponse(status_code=200, content=json.dumps(content_0), headers={"Content-Type": "application/json"}),
            MockResponse(status_code=200, content=json.dumps(content_1), headers={"Content-Type": "application/json"}),
        ]
        self.transport.request.side_effect = responses
        all_objects = await self.object_client.list_all_objects(
            limit_per_request=2,
            order_by={ObjectOrderByEnum.created_at: OrderByOperatorEnum.asc},
            schema_id=["test"],
            deleted=False,
        )
        expected_objects = [
            ObjectMetadata(
                environment=self.environment,
                id=UUID("00000000-0000-0000-0000-000000000002"),
                name="m.json",
                created_at=dateutil_parse("2020-01-01 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                parent="/A",
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                version_id="2",
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000999")),
            ),
            ObjectMetadata(
                environment=self.environment,
                id=UUID("00000000-0000-0000-0000-000000000003"),
                name="n.json",
                created_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                modified_at=dateutil_parse("2020-01-03 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                parent="/A",
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                version_id="1",
                stage=None,
            ),
            ObjectMetadata(
                environment=self.environment,
                id=UUID("00000000-0000-0000-0000-000000000002"),
                name="o.json",
                created_at=dateutil_parse("2020-01-01 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                modified_at=dateutil_parse("2020-01-02 01:30:00+00:00").replace(tzinfo=datetime.timezone.utc),
                modified_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name=None,
                    email=None,
                ),
                parent="/B",
                schema_id=ObjectSchema("objects", "test", SchemaVersion(1, 2, 3)),
                version_id="3",
                stage=None,
            ),
        ]
        self.assertEqual(expected_objects, all_objects)
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects?limit=2&offset=0&deleted=False&order_by=asc%3Acreated_at&schema_id=test",
            headers={"Accept": "application/json"},
        )
        self.assert_any_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects?limit=2&offset=2&deleted=False&order_by=asc%3Acreated_at&schema_id=test",
            headers={"Accept": "application/json"},
        )

    async def test_list_versions_by_path(self) -> None:
        object_path = "A/m.json"
        with self.transport.set_http_response(200, MOCK_VERSION_CONTENT, headers={"Content-Type": "application/json"}):
            versions = await self.object_client.list_versions_by_path(object_path)
        expected_versions = [
            ObjectVersion(
                version_id="2022-01-01T01:30:00.0000000Z",
                created_at=datetime.datetime(2022, 1, 1, 1, 30, tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name="x z",
                    email="test@example.com",
                ),
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000123")),
            ),
            ObjectVersion(
                version_id="2020-01-01T01:30:00.0000000Z",
                created_at=datetime.datetime(2020, 1, 1, 1, 30, tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000123")),
            ),
            ObjectVersion(
                version_id="2010-01-01T01:30:00.0000000Z",
                created_at=datetime.datetime(2010, 1, 1, 1, 30, tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000012"),
                    name="x w",
                    email="test@example.com",
                ),
                stage=None,
            ),
        ]

        self.assertEqual(versions, expected_versions)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/path/A/m.json?include_versions=True",
            headers={"Authorization": "Bearer <not-a-real-token>", "Accept": "application/json"},
        )

    async def test_list_versions_by_id(self) -> None:
        object_id = UUID(int=2)
        with self.transport.set_http_response(200, MOCK_VERSION_CONTENT, headers={"Content-Type": "application/json"}):
            versions = await self.object_client.list_versions_by_id(object_id)
        expected_versions = [
            ObjectVersion(
                version_id="2022-01-01T01:30:00.0000000Z",
                created_at=datetime.datetime(2022, 1, 1, 1, 30, tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000011"),
                    name="x z",
                    email="test@example.com",
                ),
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000123")),
            ),
            ObjectVersion(
                version_id="2020-01-01T01:30:00.0000000Z",
                created_at=datetime.datetime(2020, 1, 1, 1, 30, tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000010"),
                    name="x y",
                    email="test@example.com",
                ),
                stage=Stage(name="Approved", id=UUID("00000000-0000-0000-0000-000000000123")),
            ),
            ObjectVersion(
                version_id="2010-01-01T01:30:00.0000000Z",
                created_at=datetime.datetime(2010, 1, 1, 1, 30, tzinfo=datetime.timezone.utc),
                created_by=ServiceUser(
                    id=UUID("00000000-0000-0000-0000-000000000012"),
                    name="x w",
                    email="test@example.com",
                ),
                stage=None,
            ),
        ]

        self.assertEqual(expected_versions, versions)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/00000000-0000-0000-0000-000000000002?include_versions=True",
            headers={"Authorization": "Bearer <not-a-real-token>", "Accept": "application/json"},
        )

    async def test_prepare_data_upload(self) -> None:
        """Test preparing a single data upload."""
        put_data_response = load_test_data("put_data.json")
        expected_name = put_data_response[0]["name"]
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(put_data_response), headers={"Content-Type": "application/json"}
        ):
            (upload,) = [upload async for upload in self.object_client.prepare_data_upload([expected_name])]

        self.assertIsInstance(upload, ObjectDataUpload)
        self.assertEqual(expected_name, upload.name)
        self.assertEqual(self.environment, upload.environment)
        self.assert_request_made(
            method=RequestMethod.PUT,
            path=f"{self.base_path}/data",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=[{"name": expected_name}],
        )

    async def test_prepare_data_upload_batches(self) -> None:
        """Test preparing multiple data uploads in batches."""
        put_data_response = load_test_data("put_data_batch.json")
        # The service can only generate _MAX_UPLOAD_URLS upload URLs in a single request.
        batch_1 = put_data_response[:_MAX_UPLOAD_URLS]
        batch_2 = put_data_response[_MAX_UPLOAD_URLS:]

        batch_1_by_name = {
            upload["name"]: upload
            for upload in batch_1
            if upload["exists"] is False  # Pre-existing data is skipped.
        }
        batch_2_by_name = {
            upload["name"]: upload
            for upload in batch_2
            if upload["exists"] is False  # Pre-existing data is skipped.
        }

        aiter_uploads = self.object_client.prepare_data_upload([data["name"] for data in put_data_response])
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(batch_1), headers={"Content-Type": "application/json"}
        ):
            self.transport.assert_no_requests()

            # Awaiting the first result from the first batch should trigger the first request.
            upload = await anext(aiter_uploads)

            self.assert_request_made(
                method=RequestMethod.PUT,
                path=f"{self.base_path}/data",
                headers={"Content-Type": "application/json", "Accept": "application/json"},
                body=[{"name": data["name"]} for data in batch_1],
            )
            self.assertIsInstance(upload, ObjectDataUpload)
            expected = batch_1_by_name.pop(upload.name)
            self.assertEqual(expected["name"], upload.name)
            self.assertEqual(self.environment, upload.environment)

        self.transport.reset_mock()
        while len(batch_1_by_name) > 0:
            upload = await anext(aiter_uploads)
            self.assertIsInstance(upload, ObjectDataUpload)
            expected = batch_1_by_name.pop(upload.name)
            self.assertEqual(expected["name"], upload.name)
            self.assertEqual(self.environment, upload.environment)

        # No more requests should be made until the next batch is requested.
        self.transport.assert_no_requests()

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(batch_2), headers={"Content-Type": "application/json"}
        ):
            upload = await anext(aiter_uploads)
            self.assert_request_made(
                method=RequestMethod.PUT,
                path=f"{self.base_path}/data",
                headers={"Content-Type": "application/json", "Accept": "application/json"},
                body=[{"name": data["name"]} for data in batch_2],
            )
            self.assertIsInstance(upload, ObjectDataUpload)
            expected = batch_2_by_name.pop(upload.name)
            self.assertEqual(expected["name"], upload.name)
            self.assertEqual(self.environment, upload.environment)

        self.transport.reset_mock()
        while len(batch_2_by_name) > 0:
            upload = await anext(aiter_uploads)
            self.assertIsInstance(upload, ObjectDataUpload)
            expected = batch_2_by_name.pop(upload.name)
            self.assertEqual(expected["name"], upload.name)
            self.assertEqual(self.environment, upload.environment)

        # No more uploads should be available.
        with self.assertRaises(StopAsyncIteration):
            await anext(aiter_uploads)

        # No more requests should be made after all uploads are returned.
        self.transport.assert_no_requests()

    async def test_prepare_data_download(self) -> None:
        """Test preparing a single data download."""
        get_object_response = load_test_data("get_object.json")
        expected_id = UUID(get_object_response["object_id"])
        expected_version = get_object_response["version_id"]
        expected_name = get_object_response["links"]["data"][0]["name"]

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_object_response), headers={"Content-Type": "application/json"}
        ):
            (download,) = [
                download
                async for download in self.object_client.prepare_data_download(
                    expected_id, expected_version, [expected_name]
                )
            ]

        self.assertIsInstance(download, ObjectDataDownload)
        self.assertEqual(expected_name, download.name)
        self.assertEqual(expected_id, download.metadata.id)
        self.assertEqual(expected_version, download.metadata.version_id)
        self.assertEqual(self.environment, download.metadata.environment)
        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{expected_id}?version={quote(expected_version)}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )

    async def test_prepare_data_download_multiple(self) -> None:
        """Test preparing multiple data downloads."""
        get_object_response = load_test_data("get_object.json")
        expected_id = UUID(get_object_response["object_id"])
        expected_version = get_object_response["version_id"]
        expected_names = [data["name"] for data in get_object_response["links"]["data"]]
        expected_data_by_name = {data["name"]: data for data in get_object_response["links"]["data"]}

        aiter_downloads = self.object_client.prepare_data_download(expected_id, expected_version, expected_names)
        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_object_response), headers={"Content-Type": "application/json"}
        ):
            self.transport.assert_no_requests()

            # Awaiting the first result should trigger the first request.
            download = await anext(aiter_downloads)

            self.assert_request_made(
                method=RequestMethod.GET,
                path=f"{self.base_path}/objects/{expected_id}?version={quote(expected_version)}",
                headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
            )
            self.assertIsInstance(download, ObjectDataDownload)
            expected = expected_data_by_name.pop(download.name)
            self.assertEqual(expected["name"], download.name)
            self.assertEqual(expected_id, download.metadata.id)
            self.assertEqual(expected_version, download.metadata.version_id)
            self.assertEqual(self.environment, download.metadata.environment)

        self.transport.reset_mock()

        while len(expected_data_by_name) > 0:
            download = await anext(aiter_downloads)
            self.assertIsInstance(download, ObjectDataDownload)
            expected = expected_data_by_name.pop(download.name)
            self.assertEqual(expected["name"], download.name)
            self.assertEqual(expected_id, download.metadata.id)
            self.assertEqual(expected_version, download.metadata.version_id)
            self.assertEqual(self.environment, download.metadata.environment)

        # No more downloads should be available.
        with self.assertRaises(StopAsyncIteration):
            await anext(aiter_downloads)

        # No more requests should be made after all downloads are returned.
        self.transport.assert_no_requests()

    async def test_prepare_data_download_missing_data(self) -> None:
        """Test preparing to download missing data."""
        get_object_response = load_test_data("get_object.json")
        expected_id = UUID(get_object_response["object_id"])
        expected_version = get_object_response["version_id"]

        with self.transport.set_http_response(
            status_code=200, content=json.dumps(get_object_response), headers={"Content-Type": "application/json"}
        ):
            aiter_downloads = self.object_client.prepare_data_download(expected_id, expected_version, ["missing"])

            with self.assertRaises(DataNotFoundError):
                await anext(aiter_downloads)

    def test_get_data_client(self) -> None:
        """Test getting a data client."""
        data_client = self.object_client.get_data_client(self.cache)
        self.assertIsInstance(data_client, ObjectDataClient)

    def test_get_data_client_missing_dependencies(self) -> None:
        """Test getting a data client with missing dependencies."""
        with UnloadModule("evo.objects.client.api_client", "evo.objects.utils.data"), NoImport("pyarrow"):
            from evo.objects.client import ObjectAPIClient

            client = ObjectAPIClient(self.environment, self.connector)
            self.assertFalse(
                any(
                    (
                        hasattr(ObjectAPIClient, "get_data_client"),
                        hasattr(client, "get_data_client"),
                    )
                ),
                "get_data_client should not be available if pyarrow is missing",
            )

    async def test_get_latest_object_versions(self) -> None:
        content = json.dumps(
            [
                {"object_id": str(UUID(int=1)), "version_id": "2023-01-01T00:00:00.00000000Z"},
                {"object_id": str(UUID(int=2)), "version_id": None},
            ]
        )
        object_ids = [UUID(int=1), UUID(int=2)]
        with self.transport.set_http_response(200, content, headers={"Content-Type": "application/json"}):
            versions = await self.object_client.get_latest_object_versions(object_ids)
        self.assertEqual(versions, {UUID(int=1): "2023-01-01T00:00:00.00000000Z", UUID(int=2): None})
        self.assert_request_made(
            method=RequestMethod.PATCH,
            path=f"{self.base_path}/objects",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
            body=["00000000-0000-0000-0000-000000000001", "00000000-0000-0000-0000-000000000002"],
        )

    async def test_get_latest_object_versions_batch(self) -> None:
        version_id = "2023-01-01T00:00:00.00000000Z"
        object_ids = [UUID(int=i) for i in range(10)]
        batch_ids = [object_ids[i : i + 3] for i in range(0, len(object_ids), 3)]
        responses = []
        for batch in batch_ids:
            data = json.dumps([{"object_id": str(batch_id), "version_id": version_id} for batch_id in batch])
            responses.append(MockResponse(status_code=200, headers={"Content-Type": "application/json"}, content=data))
        self.transport.request.side_effect = responses
        versions = await self.object_client.get_latest_object_versions(object_ids, batch_size=3)
        expected_versions = {object_id: version_id for object_id in object_ids}
        self.assertEqual(versions, expected_versions)
        for batch in batch_ids:
            self.assert_any_request_made(
                method=RequestMethod.PATCH,
                path=f"{self.base_path}/objects",
                headers={
                    "Accept": "application/json",
                    "Content-Type": "application/json",
                },
                body=[str(b_id) for b_id in batch],
            )

    async def test_create_geoscience_object(self) -> None:
        get_object_response = load_test_data("get_object.json")
        new_pointset = {
            "name": "Sample pointset",
            "uuid": None,
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }
        new_pointset_without_uuid = new_pointset.copy()
        with self.transport.set_http_response(status_code=201, content=json.dumps(get_object_response)):
            expected_object_path = "A/m.json"
            new_object_metadata = await self.object_client.create_geoscience_object(expected_object_path, new_pointset)

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/path/{expected_object_path}",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=new_pointset_without_uuid,
        )
        expected_uuid = UUID(int=2)
        self.assertEqual("A", new_object_metadata.parent)
        self.assertEqual("m.json", new_object_metadata.name)
        self.assertEqual(expected_uuid, new_pointset["uuid"])
        self.assertEqual(expected_uuid, new_object_metadata.id)
        self.assertEqual("2023-08-03T05:47:18.3402289Z", new_object_metadata.version_id)

    async def test_create_geoscience_object_already_exists(self) -> None:
        already_exists_response = load_test_data("object_already_exists_error.json")
        expected_type = already_exists_response["type"]
        expected_title = already_exists_response["title"]
        expected_detail = already_exists_response["detail"]
        expected_existing_id = UUID(already_exists_response["existing_id"])
        expected_object_path = already_exists_response["object_path"]
        new_pointset = {
            "name": "Sample pointset",
            "uuid": None,
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }
        with (
            self.transport.set_http_response(
                status_code=409, content=json.dumps(already_exists_response), reason="Conflict"
            ),
            self.assertRaises(ObjectAlreadyExistsError) as cm,
        ):
            await self.object_client.create_geoscience_object(expected_object_path, new_pointset)

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/path/{expected_object_path}",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=new_pointset,
        )

        self.assertEqual(expected_type, cm.exception.type_)
        self.assertEqual(expected_title, cm.exception.title)
        self.assertEqual(expected_detail, cm.exception.detail)
        self.assertEqual(expected_existing_id, cm.exception.existing_id)
        self.assertEqual(expected_object_path, cm.exception.object_path)

        expected_error_str = (
            f"Error: (409) Conflict"
            f"\nType: {expected_type}"
            f"\nTitle: {expected_title}"
            f"\nDetail: {expected_detail}"
            f"\nPath: {expected_object_path}"
            f"\nExisting ID: {expected_existing_id}"
        )
        actual_error_str = str(cm.exception)
        self.assertEqual(expected_error_str, actual_error_str)

    async def test_create_geoscience_object_with_uuid_fails(self) -> None:
        existing_pointset = {
            "name": "Sample pointset",
            "uuid": "00000000-0000-0000-0000-000000000002",
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }
        with self.assertRaises(ObjectUUIDError):
            await self.object_client.create_geoscience_object("A/m.json", existing_pointset)
        self.transport.assert_no_requests()

    async def test_move_geoscience_object(self) -> None:
        get_object_response = load_test_data("get_object.json")
        existing_uuid = UUID(int=2)
        existing_pointset = {
            "name": "Sample pointset",
            "uuid": "00000000-0000-0000-0000-000000000002",
            "description": "A sample pointset object",
            "bounding_box": {
                "min_x": 0.0,
                "max_x": 0.0,
                "min_y": 0.0,
                "max_y": 0.0,
                "min_z": 0.0,
                "max_z": 0.0,
            },
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }

        with self.transport.set_http_response(status_code=201, content=json.dumps(get_object_response)):
            expected_object_path = "A/m.json"
            new_object_metadata = await self.object_client.move_geoscience_object(
                expected_object_path, existing_pointset
            )

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/path/{expected_object_path}",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=existing_pointset,
        )
        self.assertEqual("A", new_object_metadata.parent)
        self.assertEqual("m.json", new_object_metadata.name)
        self.assertEqual(existing_uuid, new_object_metadata.id)
        self.assertEqual("2023-08-03T05:47:18.3402289Z", new_object_metadata.version_id)

    async def test_move_geoscience_object_already_exists(self) -> None:
        already_exists_response = load_test_data("object_already_exists_error.json")
        expected_type = already_exists_response["type"]
        expected_title = already_exists_response["title"]
        expected_detail = already_exists_response["detail"]
        expected_existing_id = UUID(already_exists_response["existing_id"])
        expected_object_path = already_exists_response["object_path"]
        existing_pointset = {
            "name": "Sample pointset",
            "uuid": str(UUID(int=2)),
            "description": "A sample pointset object",
            "bounding_box": {
                "min_x": 0.0,
                "max_x": 0.0,
                "min_y": 0.0,
                "max_y": 0.0,
                "min_z": 0.0,
                "max_z": 0.0,
            },
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }
        with (
            self.transport.set_http_response(
                status_code=409,
                content=json.dumps(already_exists_response),
                reason="Conflict",
            ),
            self.assertRaises(ObjectAlreadyExistsError) as cm,
        ):
            await self.object_client.move_geoscience_object(expected_object_path, existing_pointset)

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/path/{expected_object_path}",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=existing_pointset,
        )

        self.assertEqual(expected_type, cm.exception.type_)
        self.assertEqual(expected_title, cm.exception.title)
        self.assertEqual(expected_detail, cm.exception.detail)
        self.assertEqual(expected_existing_id, cm.exception.existing_id)
        self.assertEqual(expected_object_path, cm.exception.object_path)

        expected_error_str = (
            f"Error: (409) Conflict"
            f"\nType: {expected_type}"
            f"\nTitle: {expected_title}"
            f"\nDetail: {expected_detail}"
            f"\nPath: {expected_object_path}"
            f"\nExisting ID: {expected_existing_id}"
        )
        actual_error_str = str(cm.exception)
        self.assertEqual(expected_error_str, actual_error_str)

    async def test_move_geoscience_object_without_uuid_fails(self) -> None:
        existing_pointset = {
            "name": "Sample pointset",
            "uuid": None,
            "description": "A sample pointset object",
            "bounding_box": {
                "min_x": 0.0,
                "max_x": 0.0,
                "min_y": 0.0,
                "max_y": 0.0,
                "min_z": 0.0,
                "max_z": 0.0,
            },
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }

        with self.assertRaises(ObjectUUIDError):
            expected_object_path = "A/m.json"
            await self.object_client.move_geoscience_object(expected_object_path, existing_pointset)
        self.transport.assert_no_requests()

    async def test_update_geoscience_object(self) -> None:
        get_object_response = load_test_data("get_object.json")
        existing_uuid = UUID(int=2)
        updated_pointset = {
            "name": "Sample pointset",
            "uuid": "00000000-0000-0000-0000-000000000002",
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }

        with self.transport.set_http_response(status_code=201, content=json.dumps(get_object_response)):
            new_object_metadata = await self.object_client.update_geoscience_object(updated_pointset)

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{existing_uuid}",
            headers={"Content-Type": "application/json", "Accept": "application/json"},
            body=updated_pointset,
        )
        self.assertEqual("A", new_object_metadata.parent)
        self.assertEqual("m.json", new_object_metadata.name)
        self.assertEqual(existing_uuid, UUID(updated_pointset["uuid"]))
        self.assertEqual(existing_uuid, new_object_metadata.id)
        self.assertEqual("2023-08-03T05:47:18.3402289Z", new_object_metadata.version_id)

    async def test_update_geoscience_object_without_uuid_fails(self) -> None:
        existing_pointset = {
            "name": "Sample pointset",
            "uuid": None,
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
        }

        with self.assertRaises(ObjectUUIDError):
            await self.object_client.update_geoscience_object(existing_pointset)
        self.transport.assert_no_requests()

    async def test_download_object_by_path(self) -> None:
        get_object_response = load_test_data("get_object.json")
        expected_uuid = UUID(int=2)
        expected_object_dict = {
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
            "uuid": UUID("00000000-0000-0000-0000-000000000002"),
            "name": "Sample pointset",
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
        }
        expected_path = "A/m.json"
        expected_version = "2023-08-03T05:47:18.3402289Z"
        with self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)):
            actual_object = await self.object_client.download_object_by_path(expected_path, expected_version)

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/path/{expected_path}?version={quote(expected_version)}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        # Check metadata.
        actual_metadata = actual_object.metadata
        self.assertEqual(expected_path, actual_metadata.path)
        self.assertEqual("A", actual_metadata.parent)
        self.assertEqual("m.json", actual_metadata.name)
        self.assertEqual(expected_uuid, actual_metadata.id)
        self.assertEqual(expected_version, actual_metadata.version_id)

        # Check geoscience object.
        self.assertEqual(expected_object_dict, actual_object.as_dict())

    async def test_download_object_by_id(self) -> None:
        get_object_response = load_test_data("get_object.json")
        expected_uuid = UUID(int=2)
        expected_object_dict = {
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
            "uuid": UUID("00000000-0000-0000-0000-000000000002"),
            "name": "Sample pointset",
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
        }
        expected_path = "A/m.json"
        expected_version = "2023-08-03T05:47:18.3402289Z"
        with self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)):
            actual_object = await self.object_client.download_object_by_id(expected_uuid, expected_version)

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/{expected_uuid}?version={quote(expected_version)}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        # Check metadata.
        actual_metadata = actual_object.metadata
        self.assertEqual(expected_path, actual_metadata.path)
        self.assertEqual("A", actual_metadata.parent)
        self.assertEqual("m.json", actual_metadata.name)
        self.assertEqual(expected_uuid, actual_metadata.id)
        self.assertEqual(expected_version, actual_metadata.version_id)

        # Check geoscience object.
        self.assertEqual(expected_object_dict, actual_object.as_dict())
        self.assertIs(actual_object.schema, actual_metadata.schema_id)

    async def test_download_object_alternate_validation(self) -> None:
        """Test downloading an object with alternate representation of confusable types.

        E.g.:
            - floats as `1` instead of `1.0`
            - integers as `1.0` instead of `1`
            - data (parquet file) ID as UUID instead of sha256 checksum
        """
        get_object_response = load_test_data("get_object_validator_check.json")
        expected_uuid = UUID(int=2)
        expected_object_dict = {
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
            "uuid": UUID("00000000-0000-0000-0000-000000000002"),
            "name": "Sample pointset",
            "description": "A sample pointset object with alternate representations of confusable types",
            "bounding_box": {"min_x": 0, "max_x": 1, "min_y": 0, "max_y": 1, "min_z": 0, "max_z": 1},
            "coordinate_reference_system": "unspecified",
            "locations": {
                "coordinates": {
                    "data": "00000000-0000-0000-0000-000000000003",
                    "length": 1.0,
                    "width": 3.0,
                    "data_type": "float64",
                }
            },
        }
        expected_path = "A/m.json"
        expected_version = "2023-08-03T05:47:18.3402289Z"
        with self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)):
            actual_object = await self.object_client.download_object_by_path(expected_path, expected_version)

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.base_path}/objects/path/{expected_path}?version={quote(expected_version)}",
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        # Check metadata.
        actual_metadata = actual_object.metadata
        self.assertEqual(expected_path, actual_metadata.path)
        self.assertEqual("A", actual_metadata.parent)
        self.assertEqual("m.json", actual_metadata.name)
        self.assertEqual(expected_uuid, actual_metadata.id)
        self.assertEqual(expected_version, actual_metadata.version_id)

        # Check geoscience object.
        self.assertEqual(expected_object_dict, actual_object.as_dict())
        self.assertIs(actual_object.schema, actual_metadata.schema_id)

    async def test_delete_object_by_path(self) -> None:
        expected_path = "A/m.json"
        with self.transport.set_http_response(status_code=204):
            actual_object = await self.object_client.delete_object_by_path(expected_path)

        self.assert_request_made(
            method=RequestMethod.DELETE,
            path=f"{self.base_path}/objects/path/{expected_path}",
            headers={"Accept-Encoding": "gzip"},
        )
        self.assertIsNone(actual_object)

    async def test_delete_object_by_id(self) -> None:
        expected_uuid = UUID(int=2)
        with self.transport.set_http_response(status_code=204):
            actual_object = await self.object_client.delete_object_by_id(expected_uuid)

        self.assert_request_made(
            method=RequestMethod.DELETE,
            path=f"{self.base_path}/objects/{expected_uuid}",
            headers={"Accept-Encoding": "gzip"},
        )
        self.assertIsNone(actual_object)

    async def test_restore_geoscience_object(self) -> None:
        expected_uuid = UUID(int=2)
        with self.transport.set_http_response(status_code=204):
            result = await self.object_client.restore_geoscience_object(expected_uuid)
            # The service returns no content on a successful restore without rename.
            assert result is None

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{expected_uuid}?deleted=False",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
        )

    async def test_restore_geoscience_object_with_rename(self) -> None:
        get_object_response = load_test_data("get_object.json")
        expected_uuid = UUID(int=2)
        # Given a server response that is a 303 redirect with the updated (post-rename) object metadata...
        with self.transport.set_http_response(status_code=303, content=json.dumps(get_object_response)):
            # ...the restored object metadata should be returned.
            restored_object_metadata = await self.object_client.restore_geoscience_object(expected_uuid)
            assert restored_object_metadata is not None

        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{self.base_path}/objects/{expected_uuid}?deleted=False",
            headers={
                "Accept": "application/json",
                "Content-Type": "application/json",
            },
        )
        self.assertEqual("A", restored_object_metadata.parent)
        self.assertEqual("m.json", restored_object_metadata.name)
        self.assertEqual(expected_uuid, restored_object_metadata.id)
        self.assertEqual("2023-08-03T05:47:18.3402289Z", restored_object_metadata.version_id)

    async def test_list_stages(self) -> None:
        list_stages_response = load_test_data("list_stages.json")

        with self.transport.set_http_response(
            200, json.dumps(list_stages_response), headers={"Content-Type": "application/json"}
        ):
            stages = await self.object_client.list_stages()

        self.assert_request_made(
            method=RequestMethod.GET,
            path=f"{self.instance_base_path}/stages",
            headers={"Accept": "application/json"},
        )

        self.assertListEqual(
            [
                Stage(id=UUID(int=1), name="Approved"),
                Stage(id=UUID(int=2), name="Experimental"),
                Stage(id=UUID(int=3), name="In Review"),
                Stage(id=UUID(int=4), name="Peer Review"),
                Stage(id=UUID(int=5), name="Preliminary Update"),
                Stage(id=UUID(int=6), name="Resource Ready"),
            ],
            stages,
        )

    async def test_set_stage(self) -> None:
        object_id = UUID(int=0)
        version_id = 1
        stage_id = UUID(int=2)

        with self.transport.set_http_response(status_code=204):
            await self.object_client.set_stage(object_id, version_id, stage_id)

        self.assert_request_made(
            method=RequestMethod.PATCH,
            path=f"{self.base_path}/objects/{object_id}/metadata?version_id={version_id}",
            headers={"Content-Type": "application/json"},
            body={"stage_id": str(stage_id)},
        )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import dataclasses
from pathlib import PurePosixPath
from typing import overload

from evo.common import Environment, Page, ServiceUser

from ..data import ObjectMetadata, ObjectSchema, ObjectVersion, OrgObjectMetadata, Stage
from ..endpoints import models

__all__ = [
    "object_metadata",
    "org_object_metadata",
    "page_of_metadata",
    "schema",
    "stage",
    "stage_or_none",
    "user_or_none",
    "version",
    "versions",
]


def user_or_none(model: models.User | None) -> ServiceUser | None:
    """Parse a ServiceUser or None value from the generated model.

    :param model: The model returned by the generated code, or None.

    :return: The parsed ServiceUser, or None if the input model is None.
    """
    return None if model is None else ServiceUser.from_model(model)


def stage(model: models.StageResponse) -> Stage:
    """Parse a Stage from the generated model.

    :param model: The model returned by the generated code.

    :return: A Stage instance.
    """
    return Stage.from_model(model)


def stage_or_none(model: models.StageResponse | None) -> Stage | None:
    """Parse a Stage or None value from the generated model.

    :param model: The model returned by the generated code, or None.

    :return: The parsed Stage, or None if the input model is None.
    """
    return None if model is None else stage(model)


def version(model: models.GeoscienceObjectVersion) -> ObjectVersion:
    """Parse an ObjectVersion from the generated model.

    :param model: The model returned by the generated code.

    :return: An ObjectVersion instance.
    """
    return ObjectVersion(
        version_id=model.version_id,
        created_at=model.created_at,
        created_by=user_or_none(model.created_by),
        stage=stage_or_none(model.stage),
    )


def versions(model: models.GetObjectResponse) -> list[ObjectVersion]:
    """Parse a list of ObjectVersion from the generated model.

    :param model: The model returned by the generated code.

    :return: A list of ObjectVersion instances, sorted by created_at in descending order.
    """
    object_versions = [version(model) for model in model.versions]
    return sorted(object_versions, key=lambda v: v.created_at, reverse=True)


def schema(schema_id: str) -> ObjectSchema:
    """Parse an ObjectSchema from the schema ID.

    :param schema_id: The schema ID string.

    :return: An ObjectSchema instance.
    """
    return ObjectSchema.from_id(schema_id)


def object_metadata(
    model: models.ListedObject | models.GetObjectResponse | models.PostObjectResponse, environment: Environment
) -> ObjectMetadata:
    """Parse an ObjectMetadata from the generated model.

    :param model: The model returned by the generated code.
    :param environment: The Evo environment associated with the object.

    :return: An ObjectMetadata instance.
    """
    # There appears to be a schema defect where object_id may possibly be None, even though it shouldn't be.
    assert model.object_id is not None

    # Parse name, parent, and schema_id from the appropriate fields depending on the model type.
    if isinstance(model, models.ListedObject):
        name = model.name
        parent = model.path.rstrip("/")
        schema_id = model.schema_
    elif model.object_path is not None:
        path = PurePosixPath(model.object_path)
        name = path.name
        parent = str(path.parent)
        schema_id = model.object.schema_
    else:
        # There appears to be _another_ schema defect where object_path may be None in
        # GetObjectResponse or PostObjectResponse, even though this never happens in practice.
        raise ValueError("Model must be a ListedObject or have an object_path")

    return ObjectMetadata(
        environment=environment,
        id=model.object_id,
        name=name,
        created_at=model.created_at,
        created_by=user_or_none(model.created_by),
        modified_at=model.modified_at,
        modified_by=user_or_none(model.modified_by),
        parent=parent,
        schema_id=schema(schema_id),
        version_id=model.version_id,
        stage=stage_or_none(model.stage),
    )


def org_object_metadata(model: models.OrgListedObject, environment: Environment) -> OrgObjectMetadata:
    """Parse an OrgObjectMetadata from the generated model.

    :param model: The model returned by the generated code.
    :param environment: The Evo environment associated with the object.

    :return: An ObjectMetadata instance.
    """
    return OrgObjectMetadata(
        environment=dataclasses.replace(environment, workspace_id=model.workspace_id),
        workspace_id=model.workspace_id,
        workspace_name=model.workspace_name,
        id=model.object_id,
        name=model.name,
        created_at=model.created_at,
        created_by=user_or_none(model.created_by),
        modified_at=model.modified_at,
        modified_by=user_or_none(model.modified_by),
        schema_id=schema(model.schema_),
        stage=stage_or_none(model.stage),
    )


@overload
def page_of_metadata(model: models.ListObjectsResponse, environment: Environment) -> Page[ObjectMetadata]: ...


@overload
def page_of_metadata(model: models.ListOrgObjectsResponse, environment: Environment) -> Page[OrgObjectMetadata]: ...


def page_of_metadata(
    model: models.ListObjectsResponse | models.ListOrgObjectsResponse, environment: Environment
) -> Page[ObjectMetadata] | Page[OrgObjectMetadata]:
    """Parse a Page of ObjectMetadata or OrgObjectMetadata from the generated model.

    :param model: The model returned by the generated code.
    :param environment: The Evo environment associated with the objects.

    :return: A Page of ObjectMetadata or OrgObjectMetadata instances.

    :raises TypeError: If the model type is unsupported.
    """
    match model:
        case models.ListObjectsResponse():
            parse_metadata = object_metadata
        case models.ListOrgObjectsResponse():
            parse_metadata = org_object_metadata
        case _:
            raise TypeError(f"Unsupported model type: {type(model)}")

    return Page(
        offset=model.offset,
        limit=model.limit,
        total=model.total,
        items=[parse_metadata(item, environment) for item in model.objects],
    )



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import contextlib
import copy
import json
from collections.abc import Generator
from typing import Any, cast
from unittest import mock
from urllib.parse import quote
from uuid import UUID

import numpy as np
import pandas as pd
import pyarrow as pa
from numpy.testing import assert_array_equal
from pandas.testing import assert_frame_equal
from parameterized import parameterized

from data import load_test_data
from evo.common import RequestMethod
from evo.common.test_tools import (
    BASE_URL,
    ORG,
    WORKSPACE_ID,
    DownloadRequestHandler,
    MultiDownloadRequestHandler,
    TestWithConnector,
    TestWithStorage,
)
from evo.common.utils import NoFeedback
from evo.jmespath import JMESPathObjectProxy
from evo.objects import DownloadedObject, ObjectReference
from evo.objects.endpoints import models
from evo.objects.exceptions import ObjectModifiedError
from evo.objects.io import _CACHE_SCOPE
from evo.objects.parquet import TableInfo
from evo.objects.utils import KnownTableFormat
from helpers import NoImport, UnloadModule, assign_property, get_sample_table_and_bytes, write_table_to_bytes

_OBJECTS_URL = f"{BASE_URL.rstrip('/')}/geoscience-object/orgs/{ORG.id}/workspaces/{WORKSPACE_ID}/objects"

_TABLE_INFO_VARIANTS: list[tuple[str, TableInfo | str]] = [
    (
        "with TableInfo dict",
        {
            "data": "0000000000000000000000000000000000000000000000000000000000000000",
            "length": 123,
            "width": 3,
            "data_type": "float64",
        },
    ),
    ("with JMESPath reference", "locations.coordinates"),
]


_category_dtype = pd.CategoricalDtype(categories=["NULL", "A", "B", "C"], ordered=False)


class TestDownloadedObject(TestWithConnector, TestWithStorage):
    def setUp(self) -> None:
        from evo.objects.client import parse

        TestWithConnector.setUp(self)
        TestWithStorage.setUp(self)

        raw = models.GetObjectResponse.model_validate(load_test_data("get_object_detailed.json"))
        self.object = DownloadedObject(
            object_=raw.object,
            metadata=parse.object_metadata(raw, self.environment),
            urls_by_name={link.name: link.download_url for link in raw.links.data},
            connector=self.connector,
            cache=self.cache,
        )

    def tearDown(self) -> None:
        # Clear cache between tests to avoid cached files interfering with subsequent tests.
        self.cache.clear_cache()

    @parameterized.expand(
        [
            ("by id as string", f"{_OBJECTS_URL}/00000000-0000-0000-0000-000000000002"),
            (
                "by id as ObjectReference",
                ObjectReference(f"{_OBJECTS_URL}/00000000-0000-0000-0000-000000000002"),
            ),
            (
                "by id with version id",
                ObjectReference(
                    f"{_OBJECTS_URL}/00000000-0000-0000-0000-000000000002?version=2023-08-03T05:47:18.3402289Z"
                ),
            ),
            ("by path as string", f"{_OBJECTS_URL}/path/A/m.json"),
            ("by path as ObjectReference", ObjectReference(f"{_OBJECTS_URL}/path/A/m.json")),
            (
                "by path with version id",
                ObjectReference(f"{_OBJECTS_URL}/path/A/m.json?version=2023-08-03T05:47:18.3402289Z"),
            ),
        ]
    )
    async def test_from_reference(self, _label: str, reference: str) -> None:
        """Test downloading a geoscience object by reference."""
        get_object_response = load_test_data("get_object.json")
        expected_uuid = UUID(int=2)
        expected_object_dict = {
            "schema": "/objects/pointset/1.0.1/pointset.schema.json",
            "uuid": UUID("00000000-0000-0000-0000-000000000002"),
            "name": "Sample pointset",
            "description": "A sample pointset object",
            "bounding_box": {"min_x": 0.0, "max_x": 0.0, "min_y": 0.0, "max_y": 0.0, "min_z": 0.0, "max_z": 0.0},
            "coordinate_reference_system": {"epsg_code": 2048},
            "locations": {
                "coordinates": {
                    "data": "0000000000000000000000000000000000000000000000000000000000000001",
                    "length": 1,
                    "width": 3,
                    "data_type": "float64",
                }
            },
        }
        expected_path = "A/m.json"
        expected_version = "2023-08-03T05:47:18.3402289Z"
        with self.transport.set_http_response(status_code=200, content=json.dumps(get_object_response)):
            actual_object = await DownloadedObject.from_reference(self.connector, reference, self.cache)

        ref = ObjectReference(reference)
        if ref.object_id is not None:
            expected_request_path = f"{_OBJECTS_URL}/{ref.object_id}"
        else:
            expected_request_path = f"{_OBJECTS_URL}/path/{ref.object_path}"

        if ref.version_id is not None:
            expected_request_path += f"?version={quote(ref.version_id)}"

        self.assert_request_made(
            method=RequestMethod.GET,
            path=expected_request_path,
            headers={"Accept": "application/json", "Accept-Encoding": "gzip"},
        )
        # Check metadata.
        actual_metadata = actual_object.metadata
        self.assertEqual(expected_path, actual_metadata.path)
        self.assertEqual("A", actual_metadata.parent)
        self.assertEqual("m.json", actual_metadata.name)
        self.assertEqual(expected_uuid, actual_metadata.id)
        self.assertEqual(expected_version, actual_metadata.version_id)

        # Check geoscience object.
        self.assertEqual(expected_object_dict, actual_object.as_dict())

    def test_search(self) -> None:
        """Test the JMESPath search implementation."""
        expected_result = JMESPathObjectProxy(
            {
                "x": [0.0, 1.0],
                "y": [2.0, 3.0],
                "z": [4.0, 5.0],
            }
        )
        actual_result = self.object.search("bounding_box | {x: [min_x, max_x], y: [min_y, max_y], z: [min_z, max_z]}")
        self.assertEqual(expected_result, actual_result)

    @parameterized.expand(
        [
            ("pass UUID", True, False),
            ("omit UUID", False, False),
            ("pass UUID with conflict check", True, True),
        ]
    )
    async def test_update(self, _label: str, pass_uuid: bool, check_for_conflict: bool) -> None:
        """Test updating a geoscience object succeeds."""
        post_object_response = load_test_data("get_object.json")
        post_object_response["version_id"] = "2"

        updated_pointset = post_object_response["object"]

        updated_pointset_parameter = copy.deepcopy(updated_pointset)
        if not pass_uuid:
            del updated_pointset_parameter["uuid"]

        with self.transport.set_http_response(status_code=201, content=json.dumps(post_object_response)):
            new_object = await self.object.update(updated_pointset_parameter, check_for_conflict=check_for_conflict)

        headers = {"Accept": "application/json", "Content-Type": "application/json"}
        if check_for_conflict:
            headers["If-Match"] = self.object.metadata.version_id
        self.assert_request_made(
            method=RequestMethod.POST,
            path=f"{_OBJECTS_URL}/{self.object.metadata.id}",
            headers=headers,
            body=updated_pointset,
        )
        # Check metadata.
        actual_metadata = new_object.metadata
        self.assertEqual(self.object.metadata.id, actual_metadata.id)
        self.assertEqual("2", actual_metadata.version_id)

        self.maxDiff = None

        # Check geoscience object.
        actual_pointset = new_object.as_dict()
        # Convert UUID in actual_pointset to a string, to make it consistent with updated_pointset
        actual_pointset["uuid"] = str(actual_pointset["uuid"])
        self.assertEqual(updated_pointset, actual_pointset)

    async def test_update_wrong_uuid(self):
        """Test updating a geoscience object fails when the object ID in the new object does not match the current object ID."""
        updated_pointset = load_test_data("get_object.json")["object"]
        updated_pointset["uuid"] = "00000000-0000-0000-0000-000000000003"
        with self.assertRaises(ValueError, msg="The object ID in the new object does not match the current object ID"):
            await self.object.update(updated_pointset)

    async def test_update_with_conflict(self):
        """Test updating a geoscience object fails when there is a new version on the server."""
        updated_pointset = load_test_data("get_object.json")["object"]
        response = load_test_data("object_modified_error.json")

        with self.transport.set_http_response(status_code=412, content=json.dumps(response)):
            with self.assertRaises(ObjectModifiedError):
                await self.object.update(updated_pointset, check_for_conflict=True)

    def _assert_optional_method(self, method_name: str, *, unload: list[str], no_import: list[str]) -> None:
        # Verify the method exists before unloading any modules.
        from evo.objects.client import DownloadedObject

        self.assertTrue(
            all(
                [
                    hasattr(DownloadedObject, method_name),
                    hasattr(self.object, method_name),
                ]
            ),
            f"DownloadedObject.{method_name} should be available for this test to be valid",
        )

        with UnloadModule("evo.objects.client.object_client", *unload), NoImport(*no_import):
            # Re-import the class to ensure the module is re-evaluated without the optional dependency.
            from evo.objects.client import DownloadedObject

            # Re-create the object to ensure the class is re-evaluated without the optional dependency.
            client = DownloadedObject(
                object_=self.object._object,
                metadata=self.object.metadata,
                urls_by_name=self.object._urls_by_name,
                connector=self.object._connector,
                cache=self.object._cache,
            )
            self.assertFalse(
                all(
                    [
                        hasattr(DownloadedObject, method_name),
                        hasattr(client, method_name),
                    ]
                ),
                f"DownloadedObject.{method_name} should not be available if "
                f"{', '.join(no_import)} {'is' if len(no_import) == 1 else 'are'} not available",
            )

    @contextlib.contextmanager
    def _patch_downloading_table(self, table_info: TableInfo | str) -> Generator[pa.Table, None, None]:
        mock_table_info = table_info
        if isinstance(mock_table_info, str):
            mock_table_info = cast(TableInfo, self.object.search(mock_table_info))

        mock_data_id = mock_table_info["data"]
        expected_filename = self.cache.get_location(self.environment, _CACHE_SCOPE) / mock_data_id
        sample_table, payload_bytes = get_sample_table_and_bytes(
            KnownTableFormat.from_table_info(mock_table_info), mock_table_info["length"]
        )
        with mock.patch("evo.common.io.download.HTTPSource", autospec=True) as mock_source:

            async def _mock_download_file_side_effect(*args, **kwargs):
                expected_download_url = self.object._urls_by_name[mock_data_id]
                actual_download_url = await kwargs["url_generator"]()
                self.assertEqual(expected_filename, kwargs["filename"])
                self.assertEqual(expected_download_url, actual_download_url)
                self.assertIs(self.transport, kwargs["transport"])
                self.assertIs(NoFeedback, kwargs["fb"])
                expected_filename.write_bytes(payload_bytes)

            mock_source.download_file.side_effect = _mock_download_file_side_effect
            yield sample_table

        mock_source.download_file.assert_called_once()
        self.transport.assert_no_requests()

    @parameterized.expand(_TABLE_INFO_VARIANTS)
    async def test_download_table(self, _label: str, table_info: TableInfo | str) -> None:
        """Test downloading parquet data as a pyarrow.Table."""
        with self._patch_downloading_table(table_info) as sample_table:
            actual_table = await self.object.download_table(table_info)

            # Should use cache second time.
            # The _patch_downloading_table context manager verifies this by checking the data is only downloaded once.
            cached_table = await self.object.download_table(table_info)

        self.assertEqual(sample_table, actual_table)
        self.assertEqual(sample_table, cached_table)

    @contextlib.contextmanager
    def _patch_downloading_table_in_memory(self, table_info: TableInfo | str) -> Generator[pa.Table, None, None]:
        self.object._cache = None  # Disable the cache for this test.

        mock_table_info = table_info
        if isinstance(mock_table_info, str):
            mock_table_info = cast(TableInfo, self.object.search(mock_table_info))

        sample_table, payload_bytes = get_sample_table_and_bytes(
            KnownTableFormat.from_table_info(mock_table_info), mock_table_info["length"]
        )

        # Use the DownloadRequestHandler from evo.common.test_tools.io to mock the binary download.
        download_handler = DownloadRequestHandler(data=payload_bytes)
        self.transport.set_request_handler(download_handler)
        yield sample_table

    @parameterized.expand(_TABLE_INFO_VARIANTS)
    async def test_download_table_without_cache(self, _label: str, table_info: TableInfo | str) -> None:
        """Test downloading parquet data in memory as a pyarrow.Table."""
        with self._patch_downloading_table_in_memory(table_info) as sample_table:
            actual_table = await self.object.download_table(table_info)

        self.assertEqual(sample_table, actual_table)

    def _set_property(self, expression: str, value: Any) -> None:
        obj = self.object.as_dict()
        assign_property(obj, expression, value)
        self.object._object = models.GeoscienceObject.model_validate(obj)

    def _setup_table(self, expression: str, table: pa.Table):
        self.object._cache = None  # Disable the cache for this test.

        # Change the number of rows in the object to match the data
        self._set_property(f"{expression}.length", table.num_rows)
        url = self.object._urls_by_name[self.object.search(f"{expression}.data")]
        payload_bytes = write_table_to_bytes(table)
        download_handler = DownloadRequestHandler(data=payload_bytes)
        self.transport.set_request_handler(download_handler)
        return url, payload_bytes

    def _setup_category(self, attribute: str):
        table = pa.Table.from_pydict(
            {
                "key": pa.array([0, 1, 2, 100], type=pa.int32()),
                "category": pa.array(["NULL", "A", "B", "C"], type=pa.string()),
            }
        )
        values = pa.Table.from_pydict(
            {
                "value": pa.array([0, 1, 2, 1, 100, 101, None], type=pa.int32()),
            }
        )

        table_url, table_bytes = self._setup_table(f"{attribute}.table", table)
        values_url, values_bytes = self._setup_table(f"{attribute}.values", values)
        self.transport.set_request_handler(
            MultiDownloadRequestHandler(
                {
                    table_url: table_bytes,
                    values_url: values_bytes,
                }
            )
        )

    async def test_download_table_nan_values(self):
        values_path = "locations.attributes[1].values"
        values = pa.Table.from_pydict(
            {
                "value": pa.array([0.2, None, 3.4, 5.3, 2.0, None], type=pa.float64()),
            }
        )
        self._setup_table(values_path, values)

        actual_table = await self.object.download_table(values_path, nan_values=[3.4, 2.0])

        expected_table = pa.Table.from_pydict(
            {
                "value": pa.array([0.2, None, None, 5.3, None, None], type=pa.float64()),
            }
        )
        self.assertEqual(actual_table, expected_table)

    async def test_download_table_column_names(self):
        with self._patch_downloading_table_in_memory("locations.coordinates") as sample_table:
            actual_table = await self.object.download_table("locations.coordinates", column_names=["XC", "YC", "ZC"])

        expected_table = sample_table.rename_columns(["XC", "YC", "ZC"])
        self.assertEqual(expected_table, actual_table)

    async def test_download_attribute_table(self):
        with self._patch_downloading_table_in_memory("locations.attributes[1].values") as sample_table:
            actual_table = await self.object.download_attribute_table("locations.attributes[1]")

        expected_table = sample_table.rename_columns(["InvRes"])
        self.assertEqual(expected_table, actual_table)

    def test_download_table_is_optional(self) -> None:
        """Test that the download_table method is not available when pyarrow is not installed."""
        self._assert_optional_method("download_table", unload=["evo.objects.parquet.loader"], no_import=["pyarrow"])

    @parameterized.expand(_TABLE_INFO_VARIANTS)
    async def test_download_dataframe(self, _label: str, table_info: TableInfo | str) -> None:
        """Test downloading parquet data as a pandas.DataFrame."""
        with self._patch_downloading_table(table_info) as sample_table:
            actual_dataframe = await self.object.download_dataframe(table_info)

            # Should use cache second time.
            # The _patch_downloading_table context manager verifies this by checking the data is only downloaded once.
            cached_dataframe = await self.object.download_dataframe(table_info)

        expected_dataframe = sample_table.to_pandas()
        assert_frame_equal(expected_dataframe, actual_dataframe)
        assert_frame_equal(expected_dataframe, cached_dataframe)

    @parameterized.expand(_TABLE_INFO_VARIANTS)
    async def test_download_dataframe_without_cache(self, _label: str, table_info: TableInfo | str) -> None:
        """Test downloading parquet data in memory as a pandas.DataFrame."""
        with self._patch_downloading_table_in_memory(table_info) as sample_table:
            actual_dataframe = await self.object.download_dataframe(table_info)

        expected_dataframe = sample_table.to_pandas()
        assert_frame_equal(expected_dataframe, actual_dataframe)

    async def test_download_dataframe_nan_values(self):
        values_path = "locations.attributes[1].values"
        values = pa.Table.from_pydict(
            {
                "value": pa.array([0.2, None, 3.4, 5.3, 2.0, None], type=pa.float64()),
            }
        )
        self._setup_table(values_path, values)

        actual_dataframe = await self.object.download_dataframe(values_path, nan_values=[0.2, 0.1, 2.0])

        expected_dataframe = pd.DataFrame({"value": [np.nan, np.nan, 3.4, 5.3, np.nan, np.nan]})
        assert_frame_equal(expected_dataframe, actual_dataframe)

    async def test_download_dataframe_column_names(self):
        with self._patch_downloading_table_in_memory("locations.coordinates") as sample_table:
            actual_dataframe = await self.object.download_dataframe(
                "locations.coordinates", column_names=["XC", "YC", "ZC"]
            )

        expected_dataframe = sample_table.to_pandas()
        expected_dataframe.columns = ["XC", "YC", "ZC"]
        assert_frame_equal(expected_dataframe, actual_dataframe)

    async def test_download_attribute_dataframe(self):
        with self._patch_downloading_table_in_memory("locations.attributes[1].values") as sample_table:
            actual_dataframe = await self.object.download_attribute_dataframe("locations.attributes[1]")

        expected_dataframe = sample_table.to_pandas()
        expected_dataframe.columns = ["InvRes"]
        assert_frame_equal(expected_dataframe, actual_dataframe)

    @parameterized.expand(
        [
            (
                "no nan values",
                None,
                None,
                pd.DataFrame({"value": ["NULL", "A", "B", "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "list of nan values",
                [0, 2],
                None,
                pd.DataFrame({"value": [pd.NA, "A", pd.NA, "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "JMESPath nan description",
                "locations.attributes[0].nan_description",
                None,
                pd.DataFrame({"value": [pd.NA, "A", "B", "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "JMESPath nan description predicate",
                "locations.attributes[?name == 'Stn'].nan_description",
                None,
                pd.DataFrame({"value": [pd.NA, "A", "B", "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "JMESPath nan value list",
                "locations.attributes[0].nan_description.values",
                None,
                pd.DataFrame({"value": [pd.NA, "A", "B", "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "JMESPath nan value list predicate",
                "locations.attributes[?name == 'Stn'].nan_description.values",
                None,
                pd.DataFrame({"value": [pd.NA, "A", "B", "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "JMESPath attribute info",
                [0, 2],
                "locations.attributes[0]",
                pd.DataFrame({"value": [pd.NA, "A", pd.NA, "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
            (
                "JMESPath attribute info predicate",
                [0, 2],
                "locations.attributes[?name == 'Stn']",
                pd.DataFrame({"value": [pd.NA, "A", pd.NA, "A", "C", pd.NA, pd.NA]}, dtype=_category_dtype),
            ),
        ]
    )
    async def test_download_category(
        self, _label: str, nan_values: str | list[int] | None, attribute_info: str | None, expected: pd.DataFrame
    ) -> None:
        attribute_path = "locations.attributes[0]"
        self._setup_category(attribute_path)

        # None, means pass it as a dictionary from the object.
        if attribute_info is None:
            attribute_info_parameter = self.object.search(attribute_path).raw
        else:
            attribute_info_parameter = attribute_info
        actual_dataframe = await self.object.download_category_dataframe(
            attribute_info_parameter, nan_values=nan_values
        )
        assert_frame_equal(expected, actual_dataframe)
        actual_table = await self.object.download_category_table(attribute_info_parameter, nan_values=nan_values)

        # Ensure the indices are int32, to be consistent with the parquet data.
        expected_table = pa.Table.from_pandas(
            expected, schema=pa.schema({"value": pa.dictionary(pa.int32(), pa.string())})
        )
        self.assertEqual(expected_table, actual_table)

        # Test loading the table through download_attribute_dataframe as well.
        if isinstance(nan_values, list):
            self._set_property(f"{attribute_path}.nan_description.values", nan_values)
        elif nan_values is None:
            self._set_property(f"{attribute_path}.nan_description.values", [])

        # None, means pass it as a dictionary from the object.
        if attribute_info is None:
            attribute_info_parameter = self.object.search(attribute_path).raw
        else:
            attribute_info_parameter = attribute_info

        expected.columns = ["Stn"]
        expected_table = expected_table.rename_columns(["Stn"])

        attribute_dataframe = await self.object.download_attribute_dataframe(attribute_info_parameter)
        assert_frame_equal(expected, attribute_dataframe)
        actual_table = await self.object.download_attribute_table(attribute_info_parameter)
        self.assertEqual(expected_table, actual_table)

    @parameterized.expand(
        [
            ("pyarrow",),
            ("pandas",),
        ]
    )
    def test_download_dataframe_is_optional(self, missing: str) -> None:
        """Test that the download_dataframe method is not available when pandas or pyarrow is not installed."""
        self._assert_optional_method("download_dataframe", unload=["evo.objects.parquet.loader"], no_import=[missing])

    @parameterized.expand(_TABLE_INFO_VARIANTS)
    async def test_download_array(self, _label: str, table_info: TableInfo | str) -> None:
        """Test downloading parquet data as a numpy.ndarray."""
        with self._patch_downloading_table(table_info) as sample_table:
            actual_array = await self.object.download_array(table_info)

            # Should use cache second time.
            # The _patch_downloading_table context manager verifies this by checking the data is only downloaded once.
            cached_array = await self.object.download_array(table_info)

        expected_array = sample_table.to_pandas().to_numpy()
        assert_array_equal(expected_array, actual_array, strict=True)
        assert_array_equal(expected_array, cached_array, strict=True)

    @parameterized.expand(_TABLE_INFO_VARIANTS)
    async def test_download_array_without_cache(self, _label: str, table_info: TableInfo | str) -> None:
        """Test downloading parquet data in memory as a numpy.ndarray."""
        with self._patch_downloading_table_in_memory(table_info) as sample_table:
            actual_array = await self.object.download_array(table_info)

        expected_array = sample_table.to_pandas().to_numpy()
        assert_array_equal(expected_array, actual_array, strict=True)

    @parameterized.expand(
        [
            ("pyarrow",),
            ("numpy",),
        ]
    )
    def test_download_array_is_optional(self, missing: str) -> None:
        """Test that the download_array method is not available when numpy or pyarrow is not installed."""
        self._assert_optional_method("download_array", unload=["evo.objects.parquet.loader"], no_import=[missing])



#  Copyright © 2025 Bentley Systems, Incorporated
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#      http://www.apache.org/licenses/LICENSE-2.0
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import unittest
from unittest import mock

from parameterized import parameterized

from evo.colormaps.data import CategoryColormap, ContinuousColormap, DiscreteColormap, _validate_colors


class TestValidateColors(unittest.TestCase):
    def test_valid_colors(self):
        colors = [[255, 240, 219], [238, 217, 196], [217, 185, 155]]
        try:
            _validate_colors(colors)
        except ValueError:
            self.fail("_validate_colors raised ValueError unexpectedly!")

    def test_valid_colours_for_category(self):
        colors = [[253, 240, 219]]
        _validate_colors(colors, minimum=1)

    def test_colors_out_of_range(self):
        colors = [[256, 240, 219], [238, 217, 196], [217, 185, 155]]
        with self.assertRaises(ValueError):
            _validate_colors(colors)

    def test_colors_incorrect_length(self):
        colors = [[255, 240], [238, 217, 196], [217, 185, 155]]
        with self.assertRaises(ValueError):
            _validate_colors(colors)

    def test_colors_too_many_entries(self):
        colors = [[255, 240, 219]] * 1025
        with self.assertRaises(ValueError):
            _validate_colors(colors)

    def test_colors_too_few_entries(self):
        colors = [[255, 240, 219]]
        with self.assertRaises(ValueError):
            _validate_colors(colors)

    def test_colors_wrong_numeric_data_type(self):
        colors = [[255, 240, 219], [238, 217, 196], [217, 185, 155], [1.0, 0.5, 0.2]]
        with self.assertRaises(ValueError):
            _validate_colors(colors)

    def test_colors_wrong_data_type(self):
        colors = [[255, 240, 219], [238, 217, 196], [217, 185, 155], ["a", "b", "c"]]
        with self.assertRaises(ValueError):
            _validate_colors(colors)

    @parameterized.expand(
        [
            ("continuous", 2, 1024, ContinuousColormap),
            ("discrete", 1, 1024, DiscreteColormap),
            ("category", 1, 10000, CategoryColormap),
        ]
    )
    def test_colormap_creation_validates(self, _, minimum_colors, maximum_colours, colormap_dataclass):
        kwargs = {param: mock.MagicMock() for param in colormap_dataclass.__dataclass_fields__ if param != "colors"}
        with mock.patch("evo.colormaps.data._validate_colors") as mock_validate_colors:
            colors = mock.MagicMock()
            colormap_dataclass(colors=colors, **kwargs)
            mock_validate_colors.assert_called_once_with(colors, minimum=minimum_colors, maximum=maximum_colours)
